#include <Padico/Module.h>
#include <Padico/Puk.h>
#include <Padico/Timing.h>

#include "dummy.h"

PADICO_MODULE_DECLARE(Test_Components);
PADICO_MODULE_INIT(test_comp_init);
PADICO_MODULE_RUN(test_comp_run);

/* ********************************************************* */
/* *** drivers */

static void*test_comp_instantiate(puk_instance_t ai, puk_context_t context);
static void test_comp_destroy(void*_instance);

const static struct puk_component_driver_s test_comp_component_driver =
  {
    .instantiate = &test_comp_instantiate,
    .destroy     = &test_comp_destroy
  };

static int test_comp_dummy(int a);

const static struct padico_dummy_driver_s test_comp_dummy_driver =
  {
    .dummy = &test_comp_dummy
  };

/* ********************************************************* */
/* *** init */

static puk_component_t component;
static puk_instance_t instance;

static int test_comp_init(void)
{
  component = 
    puk_component_declare("TestComponents",
			puk_component_provides("PadicoComponent", "component", &test_comp_component_driver),
			puk_component_provides("Dummy",         "dummy", &test_comp_dummy_driver));
  instance = puk_component_instantiate(component);
  return 0;
}

static int test_comp_run(int argc, char**argv)
{
  int i;
  const int iters = 100000;
  padico_timing_t t1, t2;
  double t;
  
  padico_print("TestComponents: *** Components ***\n");
  
  /* *** iface_lookup */
  padico_timing_get_tick(&t1);
  for(i = 0; i < iters; i++)
    {
      puk_iface_lookup("Dummy");
    }
  padico_timing_get_tick(&t2);
  t = padico_timing_diff_usec(&t1, &t2);
  padico_print("TestComponents: puk_iface_lookup: %fusec\n", t/iters);

  /* *** component_resolve */
  padico_timing_get_tick(&t1);
  for(i = 0; i < iters; i++)
    {
      puk_component_resolve("TestComponents");
    }
  padico_timing_get_tick(&t2);
  t = padico_timing_diff_usec(&t1, &t2);
  padico_print("TestComponents: puk_component_resolve: %fusec\n", t/iters);

  /* *** get_driver */
  padico_timing_get_tick(&t1);
  for(i = 0; i < iters; i++)
    {
      puk_component_get_driver_Dummy(component, NULL);
    }
  padico_timing_get_tick(&t2);
  t = padico_timing_diff_usec(&t1, &t2);
  padico_print("TestComponents: puk_component_getdriver: %fusec\n", t/iters);

  /* *** instance_indirect */
  struct puk_receptacle_Dummy_s comp_receptacle;
  padico_timing_get_tick(&t1);
  for(i = 0; i < iters; i++)
    {
      puk_instance_indirect_Dummy(instance, NULL, &comp_receptacle);
    }
  padico_timing_get_tick(&t2);
  t = padico_timing_diff_usec(&t1, &t2);
  padico_print("TestComponents: puk_instance_indirect: %fusec\n", t/iters);

  /* *** invoke: indirect, local */
  padico_timing_get_tick(&t1);
  for(i = 0; i < iters; i++)
    {
      (*comp_receptacle.driver->dummy)(1);
    }
  padico_timing_get_tick(&t2);
  t = padico_timing_diff_usec(&t1, &t2);
  padico_print("TestComponents: invoke (indirect, local): %fusec\n", t/iters);

  /* *** invoke: indirect, extern */
  puk_component_t far_component = puk_component_resolve("TestFar");
  puk_instance_t far_instance = puk_component_instantiate(far_component);
  struct puk_receptacle_Dummy_s far_dummy;
  puk_instance_indirect_Dummy(far_instance, NULL, &far_dummy);
  padico_timing_get_tick(&t1);
  for(i = 0; i < iters; i++)
    {
      (*far_dummy.driver->dummy)(1);
    }
  padico_timing_get_tick(&t2);
  t = padico_timing_diff_usec(&t1, &t2);
  padico_print("TestComponents: invoke (indirect, extern): %fusec\n", t/iters);

  /* *** invoke: direct, local */
  padico_timing_get_tick(&t1);
  for(i = 0; i < iters; i++)
    {
      test_comp_dummy(1);
    }
  padico_timing_get_tick(&t2);
  t = padico_timing_diff_usec(&t1, &t2);
  padico_print("TestComponents: invoke (direct, local): %fusec\n", t/iters);

  /* *** invoke: direct, extern */
  padico_timing_get_tick(&t1);
  for(i = 0; i < iters; i++)
    {
      test_extern_dummy(1);
    }
  padico_timing_get_tick(&t2);
  t = padico_timing_diff_usec(&t1, &t2);
  padico_print("TestComponents: invoke (direct, extern): %fusec\n", t/iters);

  /* *** iface_lookup, instance_indirect & invoke */
  padico_timing_get_tick(&t1);
  for(i = 0; i < iters; i++)
    {
      puk_instance_indirect_Dummy(instance, NULL, &comp_receptacle);
      (*comp_receptacle.driver->dummy)(1);
    }
  padico_timing_get_tick(&t2);
  t = padico_timing_diff_usec(&t1, &t2);
  padico_print("TestComponents: iface_lookup, instance_indirect & invoke: %fusec\n", t/iters);


  padico_print("TestComponents: *** Data containers ***\n");

  /* *** hash function */
  padico_timing_get_tick(&t1);
  for(i = 0; i < iters; i++)
    {
      puk_hash_int((void*)1);
    }
  padico_timing_get_tick(&t2);
  t = padico_timing_diff_usec(&t1, &t2);
  padico_print("TestComponents: puk_hash_int: %fusec\n", t/iters);

  const char hashtest[128] = {0};
  padico_timing_get_tick(&t1);
  for(i = 0; i < iters; i++)
    {
      puk_hash_oneatatime(hashtest, 16);
    }
  padico_timing_get_tick(&t2);
  t = padico_timing_diff_usec(&t1, &t2);
  padico_print("TestComponents: puk_hash_oneatatime(16): %fusec\n", t/iters);
  padico_timing_get_tick(&t1);
  for(i = 0; i < iters; i++)
    {
      puk_hash_oneatatime(hashtest, 128);
    }
  padico_timing_get_tick(&t2);
  t = padico_timing_diff_usec(&t1, &t2);
  padico_print("TestComponents: puk_hash_oneatatime(128): %fusec\n", t/iters);


  /* *** hashtable */
  puk_hashtable_t h = puk_hashtable_new_int();
  puk_hashtable_insert(h, (void*)1, instance);
  padico_timing_get_tick(&t1);
  for(i = 0; i < iters; i++)
    {
      puk_hashtable_lookup(h, (void*)1);
    }
  padico_timing_get_tick(&t2);
  t = padico_timing_diff_usec(&t1, &t2);
  padico_print("TestComponents: puk_hashtable_lookup (no load): %fusec\n", t/iters);
  for(i = 0; i < 1000; i++)
    {
      puk_hashtable_insert(h, (void*)(rand()), instance);
    }
  padico_timing_get_tick(&t1);
  for(i = 0; i < iters; i++)
    {
      puk_hashtable_lookup(h, (void*)1);
    }
  padico_timing_get_tick(&t2);
  t = padico_timing_diff_usec(&t1, &t2);
  padico_print("TestComponents: puk_hashtable_lookup (loaded): %fusec\n", t/iters);

  /* *** mutex */

  padico_print("TestComponents: *** Multi-threading ***\n");

  marcel_mutex_t mutex;
  marcel_mutex_init(&mutex, NULL);
  padico_timing_get_tick(&t1);
  for(i = 0; i < iters; i++)
    {
      marcel_mutex_lock(&mutex);
      marcel_mutex_unlock(&mutex);
    }
  padico_timing_get_tick(&t2);
  t = padico_timing_diff_usec(&t1, &t2);
  padico_print("TestComponents: marcel mutex lock+unlock: %fusec\n", t/iters);

  /* *** memcpy */
  padico_print("TestComponents: *** Memcpy ***\n");

  {
    long source, dest;
    padico_timing_get_tick(&t1);
    for(i = 0; i < iters; i++)
      {
	memcpy(&dest, &source, sizeof(source));
      }
    padico_timing_get_tick(&t2);
    t = padico_timing_diff_usec(&t1, &t2);
    padico_print("TestComponents: memcpy sizeof(long): %fusec\n", t/iters);
  }
  {
    const int msize = 128;
    char source[msize], dest[msize];
    padico_timing_get_tick(&t1);
    for(i = 0; i < iters; i++)
      {
	memcpy(dest, source, msize);
      }
    padico_timing_get_tick(&t2);
    t = padico_timing_diff_usec(&t1, &t2);
    padico_print("TestComponents: memcpy 128 bytes: %fusec\n", t/iters);
  }
  {
    const int msize = 4096;
    char source[msize], dest[msize];
    padico_timing_get_tick(&t1);
    for(i = 0; i < iters; i++)
      {
	memcpy(dest, source, msize);
      }
    padico_timing_get_tick(&t2);
    t = padico_timing_diff_usec(&t1, &t2);
    padico_print("TestComponents: memcpy 4 KB: %fusec\n", t/iters);
  }
  return 0;
}

/* ********************************************************* */
/* *** implementation */

static int test_comp_dummy(int a) __attribute__((noinline));
static int test_comp_dummy(int a)
{
  return a;
}

static void*test_comp_instantiate(puk_instance_t instance, puk_context_t context)
{
  return padico_malloc(sizeof(int));
}

static void test_comp_destroy(void*_instance)
{
  free(_instance);
}
