/** @file
 * @brief Padico NetAccess control channel over SysIO
 * @ingroup Control
 */

#include <Padico/Puk.h>
#include <Padico/Module.h>
#include <Padico/PadicoTM.h>
#include <Padico/SysIO.h>
#include <Padico/PadicoControl.h>
#include <Padico/SocketFactory.h>
#include <arpa/inet.h>

PADICO_MODULE_DECLARE(Control_SysIO);
PADICO_MODULE_INIT(control_sysio_init);

/** @defgroup Control_SysIO component: Control/SysIO- control channel over SysIO
 * @ingroup PadicoComponent
 */

/* *** PadicoControl driver ******************************** */

static void sysio_control_send_common(void*_status, padico_na_node_t node, padico_control_msg_t msg);

/** 'Control' facet for Control/SysIO
 * @ingroup Control_SysIO
 */
const static struct padico_na_control_driver_s sysio_control_driver =
{
  .send_common = &sysio_control_send_common
};


/* *** PadicoBootstrap driver ******************************** */

static int sysio_control_bootstrap_new(void*_status, const struct in_addr*ip, uint16_t*port);
static int sysio_control_bootstrap_connect(void*_status, const struct in_addr*ip, const uint16_t port);

/** 'Bootstrap' facet for Control/SysIO
 * @ingroup Control_SysIO
 */
const static struct padico_bootstrap_driver_s sysio_bootstrap_driver =
{
  .bootstrap_new     = &sysio_control_bootstrap_new,
  .bootstrap_connect = &sysio_control_bootstrap_connect
};

/* *** PadicoComponent driver ******************************** */

static void*sysio_control_instantiate(puk_instance_t, puk_context_t);
static void sysio_control_destroy(void*);

/** instanciation facet for Control/SysIO
 * @ingroup Control_SysIO
 */
const static struct puk_component_driver_s sysio_control_component_driver =
{
  .instantiate = &sysio_control_instantiate,
  .destroy     = &sysio_control_destroy
};


/* ********************************************************* */

enum sysio_control_state_e
  {
    SYSIO_CONTROL_STATE_INIT = 0,    /**< connection is initializing */
    SYSIO_CONTROL_STATE_CONNECTED,   /**< connection is established */
    SYSIO_CONTROL_STATE_DISCONNECTED /**< connection has been disconnected */
  };

/** status of an instance of a Control/SysIO component.
 * @ingroup Control_SysIO
 */
struct sysio_control_status_s
{
  puk_instance_t instance;        /**< self as an instance (used for interactions with padico_control_*) */
  struct puk_receptacle_SocketFactory_s sf;
  puk_context_t context;
  marcel_mutex_t lock;            /**< lock for the SysIO control channel */
  marcel_cond_t connection_event; /**< condition signaled upon connection event */
  enum sysio_control_state_e state;  /**< whether the channel is connected */
  padico_req_t req;               /**< request object used for asynchronous connection */
  padico_na_node_t peer;          /**< peer node */
  padico_io_t in_io;              /**< IO handler (used for receipt only) */
  int fd;                         /**< physical file descriptor of the socket (puk_abi_wrap_* level) */
};

/* ********************************************************* */
/* *** PadicoControl */

static void sysio_control_connector_notify(struct sysio_control_status_s*status, padico_rc_t rc);

static void sysio_control_connection_lost(struct sysio_control_status_s*status)
{
  status->state = SYSIO_CONTROL_STATE_DISCONNECTED;
  padico_control_delete_node(status->instance, status->peer);
  sysio_control_connector_notify(status, padico_rc_error("Control-SysIO: connection lost"));
}

static void sysio_control_send_common(void*_status, padico_na_node_t node, padico_control_msg_t msg)
{
  struct sysio_control_status_s*status = _status;
  const struct iovec*v = padico_control_msg_get_iovec(msg);
  const int iovcnt = padico_control_msg_get_iovec_size(msg);
  const size_t size = padico_control_msg_get_size(msg);
  struct iovec*v2 = padico_malloc(sizeof(struct iovec) * (iovcnt + 1));
  int i;
  for(i=0; i<iovcnt; i++)
    {
      v2[i+1].iov_base = v[i].iov_base;
      v2[i+1].iov_len = v[i].iov_len;
    }

  padico_out(30, "locking- fd=%d; size=%ld.\n", status->fd, puk_ulong(size));
  marcel_mutex_lock(&status->lock);
  while(status->state == SYSIO_CONTROL_STATE_INIT)
    {
      padico_out(80, "wait for the establishment of the control channel.\n");
      marcel_cond_wait(&status->connection_event, &status->lock);
    }
  assert(status->fd > 0);
  assert(node == status->peer);
  padico_out(30, "sending- fd=%d; size=%ld.\n", status->fd, puk_ulong(size));
  const uint32_t n_size = htonl(size);
  v2[0].iov_base = (void*)&n_size;
  v2[0].iov_len = sizeof(n_size);
  int sysrc = padico_sysio_outv(status->fd, v2, iovcnt + 1);
  marcel_mutex_unlock(&status->lock);
  if(sysrc != (size + sizeof(n_size)))
    {
      sysio_control_connection_lost(status);
    }
  padico_free(v2);
  padico_out(30, "send ok.\n");
}

static int sysio_control_read_callback(int fd, void*key)
{
  int again = 1;
  struct sysio_control_status_s*status = (struct sysio_control_status_s*)key;
  uint32_t n_size = 0;

  assert(status->fd == fd);
  padico_out(40, "entering...\n");
  int sysrc = padico_sysio_in(fd, &n_size, sizeof(uint32_t));
  if(sysrc <= 0)
    {
      again = 0;
      sysio_control_connection_lost(status);
      return again;
    }
  const size_t size = htonl(n_size);
  char*bytes = NULL;
  if(size > 0)
    {
      bytes = padico_malloc(size);
      sysrc = padico_sysio_in(fd, bytes, size);
      if(sysrc < size)
	{
	  again = 0;
	  sysio_control_connection_lost(status);
	  return again;
	}
    }
  padico_out(40, "received message- len=%ld; msg=*%s*\n", puk_ulong(size), bytes);
  padico_control_deliver_message(status->instance, status->peer, bytes, size);
  return again;
}



/* ********************************************************* */

/* *** socket handles */

/* ip and port must be in network byte order */
static int sysio_control_socket_create(struct sysio_control_status_s*status, 
				       const struct in_addr*ip, const uint16_t port,
				       struct sockaddr_in* servaddr)
{
  assert(status != NULL);
  puk_context_indirect_SocketFactory(status->instance, "sf", &status->sf);
  assert(status->sf.driver && status->sf._status);
  int rc = (*status->sf.driver->create)(status->sf._status, AF_INET, SOCK_STREAM, 0);
  if(rc < 0)
    padico_warning("Control-SysIO: socket_create: cannot create the socket ip=%s:%d.\n", 
		   inet_ntoa(servaddr->sin_addr), ntohs(port));

  memset(servaddr, 0, sizeof(*servaddr));
  servaddr->sin_family = AF_INET;
  servaddr->sin_addr.s_addr = ip->s_addr;
  servaddr->sin_port = port;

  return rc;
}


static int sysio_control_socket_connect(struct sockaddr_in*servaddr,
					padico_socketfactory_connector_t connector,
					struct sysio_control_status_s*status)
{
  if(status->sf.driver->connect == NULL)
    {
      padico_warning("Control-SysIO: boostrap_connect: "
		     "the socketfactory does not provide the required driver (connect).\n");
      return -1;
    }
  padico_out(6, "connecting to addr=%s; port=%d\n", inet_ntoa(servaddr->sin_addr), ntohs(servaddr->sin_port)); 
  int rc = (*status->sf.driver->connect)(status->sf._status, (struct sockaddr*) servaddr, sizeof(*servaddr),
					 connector, status);
  if(rc == -EINPROGRESS)
    rc = 0;
  if(rc != 0)
    padico_out(30, "socket_connect: Connection failed- rc = %d (%s).\n", rc, strerror(-rc));
  return rc;
}

static int sysio_control_socket_close(struct sysio_control_status_s*status)
{
  int rc = 0;
  assert(status != NULL);
  if(status->sf.driver && status->sf.driver->close)
    rc = (*status->sf.driver->close)(status->sf._status);
  if(status->in_io)
    padico_io_unregister(status->in_io);
  status->in_io = NULL;
  if(status->fd != -1)
    padico_sysio_close(status->fd);
  status->fd = -1;
  padico_out(40, "connection closed, rc=%d\n", rc);
  return rc;
}

/* ********************************************************* */

/* *** PadicoConnector */


static void sysio_control_connector_notify(struct sysio_control_status_s*status, padico_rc_t rc)
{
  if(status->req)
    {
      padico_tm_req_notify(status->req->id, rc);
      status->req = NULL;
    }
  else padico_rc_delete(rc);
}

static void*sysio_control_start(void*_status)
{
  struct sysio_control_status_s*status = _status;
  /* announce local node to peer node */
  padico_string_t self_node = padico_na_node_serialize(padico_na_getlocalnode());
  const char*msg = padico_string_get(self_node); 
  marcel_mutex_lock(&status->lock);
  assert(status->fd > 0);
  assert(status->state == SYSIO_CONTROL_STATE_INIT);
  uint32_t size = strlen(msg) + 1;
  int sysrc = padico_sysio_out(status->fd, &size, sizeof(size));
  if(sysrc > 0)
    {
      sysrc = padico_sysio_out(status->fd, msg, size);
    }
  marcel_mutex_unlock(&status->lock);
  if(sysrc < 0)
    {
      sysio_control_connection_lost(status);
    }
  msg = NULL;
  padico_string_delete(self_node);
  /* receive peer announce */
  sysrc = padico_sysio_in(status->fd, &size, sizeof(size));
  if(sysrc <= 0)
    {
      sysio_control_connection_lost(status);
      return NULL;
    }
  char*recv_msg = padico_malloc(size);
  sysrc = padico_sysio_in(status->fd, recv_msg, size);
  if(sysrc < size)
    {
      sysio_control_connection_lost(status);
      return NULL;
    }
  /* parse peer announce */
  padico_control_event_disable();
  struct puk_parse_entity_s e = puk_xml_parse_buffer(recv_msg, size - 1, PUK_TRUST_CLUSTER);
  padico_rc_t rc = puk_parse_get_rc(&e);
  if(padico_rc_iserror(rc))
    {
      padico_rc_show(rc);
      padico_fatal("Control-SysIO: parse error on initial handshake message.");
    }
  if(!puk_parse_is(&e, "Topology:node"))
    {
      padico_fatal("Control-SysIO: root entity is not \"Topology:node\" while initializing control channel.\n");
    }
  padico_na_node_t peer = puk_parse_get_content(&e);
  assert(peer != NULL);
  assert(status->peer == NULL);
  status->peer = peer;
  padico_control_new_node(status->instance, peer);
  padico_control_event_enable();

  assert(padico_na_getnodebyuuid(padico_na_node_getuuid(peer)) != NULL);

  /* authorize to send */
  marcel_mutex_lock(&status->lock);
  status->state = SYSIO_CONTROL_STATE_CONNECTED;
  status->in_io = padico_io_register(status->fd, PADICO_IO_EVENT_READ,
				     &sysio_control_read_callback, status);
  marcel_cond_broadcast(&status->connection_event);
  marcel_mutex_unlock(&status->lock);
  /* activate polling */
  padico_io_activate(status->in_io);

  /* notify connection establishment */
  padico_out(20, "the connection to node %s is up\n", padico_na_node_getname(peer));
  sysio_control_connector_notify(status, padico_rc_ok());
  padico_free(recv_msg);
  /* flush all events that may have been generate by the new_node event handler
   * since they may declare new nodes (from router) that will send us messages.
   * Remote routers may push messages before our local router event knows they exist!
   * Better know them *before* activating I/O to avoid troubles.
   */
  padico_control_event_disable();
  padico_control_event_flush();
  padico_control_event_enable();
 
  return NULL;
}



/* ********************************************************* */
/* *** PadicoBootstrap */

static void sysio_control_bootstrap_acceptor(void*key, int new_fd, const struct sockaddr*addr, socklen_t addrlen)
{
  struct sysio_control_status_s*server_status = (struct sysio_control_status_s*) key;
  /* the new controler will manage this new socket */
  padico_out(10, "new bootstrap connection received from=%s.\n", 
	     inet_ntoa(((struct sockaddr_in*) addr)->sin_addr));
  puk_instance_t instance = puk_component_instantiate(server_status->instance->component);
  struct puk_receptacle_PadicoControl_s r;
  puk_instance_indirect_PadicoControl(instance, NULL, &r);
  struct sysio_control_status_s*conn_status = r._status;
  conn_status->fd = new_fd; 
  padico_control_plug_controler(conn_status->instance, server_status->instance);
  padico_tm_invoke_later(&sysio_control_start, conn_status);
  padico_out(10, "bootstrap connection accepted from=%s.\n", 
	     inet_ntoa(((struct sockaddr_in*) addr)->sin_addr));
}

static void sysio_control_bootstrap_connector(void*key, int fd)
{
  struct sysio_control_status_s*status = (struct sysio_control_status_s*) key;
  padico_out(10, "bootstrap connecting...\n");
  status->fd = fd;
  padico_control_plug_controler(status->instance, NULL);
  padico_tm_invoke_later(&sysio_control_start, status);
  padico_out(10, "bootstrap connected.\n");
}

static int sysio_control_bootstrap_new(void*_status, const struct in_addr*ip, uint16_t*port) 
{ 
  padico_out(20, "entering.\n");
  struct sysio_control_status_s*status = (struct sysio_control_status_s*) _status;
  struct sockaddr_in servaddr;
  sysio_control_socket_create(status, ip, *port, &servaddr);
  const int default_backlog = 32;
  int backlog = padico_na_size() > default_backlog ? padico_na_size() : default_backlog;
  if((status->sf.driver->bind == NULL) || (status->sf.driver->listen == NULL))
    {
      padico_fatal("Control-SysIO: socket_listen: the socketfactory does not provide the required methods (bind, listen).\n");
      return -1;
    }
  if(status->sf.driver->setsockopt)
    {
      int val = 1;
      (*status->sf.driver->setsockopt)(status->sf._status, SOL_SOCKET, SO_REUSEADDR, &val, sizeof(val));
    }
  socklen_t servaddr_len; 
  int rc = (*status->sf.driver->bind)(status->sf._status, (struct sockaddr*) &servaddr, &servaddr_len);
  if(!rc)
    {
      rc = (*status->sf.driver->listen)(status->sf._status, backlog, (struct sockaddr*) &servaddr, servaddr_len,
					&sysio_control_bootstrap_acceptor, status);
    }
  if(rc)
    {
      padico_fatal("Control-SysIO: socket_listen: err %d (%s)\n", -rc, strerror(-rc));
    }
  if(*port == 0)
    *port = servaddr.sin_port;
  return rc;
}

static int sysio_control_bootstrap_connect(void*_status, const struct in_addr*ip, const uint16_t port)
{
  struct sysio_control_status_s*status = (struct sysio_control_status_s*) _status;
  struct sockaddr_in servaddr;
  sysio_control_socket_create(status, ip, port, &servaddr);
  int rc = sysio_control_socket_connect(&servaddr, &sysio_control_bootstrap_connector, status);
  if(rc == 0)
    {
      marcel_mutex_lock(&status->lock);
      while(status->state == SYSIO_CONTROL_STATE_INIT)
	{
	  padico_out(80, "wait for the establishment of the control channel.\n");
	  marcel_cond_wait(&status->connection_event, &status->lock);
	}
      if(status->state != SYSIO_CONTROL_STATE_CONNECTED)
	{
	  rc = -ENETUNREACH;
	}
      marcel_mutex_unlock(&status->lock);
    }
  return rc;
}    

/* ********************************************************* */
/* *** PadicoComponent */

static void*sysio_control_instantiate(puk_instance_t instance, puk_context_t context)
{
  struct sysio_control_status_s*status = padico_malloc(sizeof(struct sysio_control_status_s));
  *status = (struct sysio_control_status_s) 
    {
      .peer     = NULL,
      .fd       = -1,
      .in_io    = NULL,
      .state    = SYSIO_CONTROL_STATE_INIT,
      .req	= NULL,
      .instance = instance,
      .context  = context,
      .sf       = { .driver = NULL, ._status = NULL}
    };
  marcel_mutex_init(&status->lock, 0);
  marcel_cond_init(&status->connection_event, 0);
  return (void*)status;
}

static void sysio_control_destroy(void*_status)
{
  if(_status)
    {
      struct sysio_control_status_s*status = _status;
      if(status->fd != -1)
	sysio_control_socket_close(status);
      marcel_mutex_destroy(&status->lock);
      marcel_cond_destroy(&status->connection_event);
      padico_free(_status);
    }
}

/* ********************************************************* */

int control_sysio_init(void)
{
  puk_component_declare("Control-SysIO",
			puk_component_provides("PadicoComponent",   "component",   &sysio_control_component_driver), 
			puk_component_provides("PadicoControl",   "control",   &sysio_control_driver),
			puk_component_provides("PadicoBootstrap", "bootstrap", &sysio_bootstrap_driver),
			puk_component_uses("SocketFactory", "sf"));
  padico_control_event_sync();
  return 0;
}

