/** @file
 * @brief LZ4 Packet Filter component
 */

#include <Padico/Module.h>
#include "PacketFilter.h"
#include <lz4.h>

PADICO_MODULE_DECLARE(PacketFilter_LZ4);
PADICO_MODULE_INIT(lz4pf_init);

/* ********************************************************* */
/* *** LZ4 Packet Filter *********************************** */
/* ********************************************************* */


/** @defgroup PacketFilter_LZ4 component: PacketFilter-LZ4- a packet filter that performs LZ4 compression
 * @ingroup PadicoComponent
 */

static void*lz4pf_instantiate(puk_instance_t assembly, puk_context_t context);
static void lz4pf_destroy(void*state);
static void lz4pf_encode(void*, const void*, size_t, void**, size_t*);
static void lz4pf_decode(void*, const void*, size_t, void*, size_t*);

/** instanciation facet for PacketFilter-LZ4
 * @ingroup PacketFilter_LZ4
 */
const static struct puk_component_driver_s lz4_component_driver =
  {
    .instantiate = &lz4pf_instantiate,
    .destroy     = &lz4pf_destroy
  };
/** 'PacketFilter' facet for PacketFilter-LZ4
 * @ingroup PacketFilter_LZ4
 */
const static struct packetfilter_driver_s lz4pf_driver = 
  {
    .pf_encode = &lz4pf_encode,
    .pf_decode = &lz4pf_decode
  };

/** instance status for PacketFilter-LZ4
 * @ingroup PacketFilter_LZ4
 */
struct lz4_state_s
{
  int dummy;
};


static void* lz4pf_instantiate(puk_instance_t assembly, puk_context_t context)
{
  struct lz4_state_s*state = padico_malloc(sizeof(struct lz4_state_s));
  state->dummy = 0;
  return state;
}

static void lz4pf_destroy(void*state)
{
  assert(state != NULL);
  padico_free(state);
}


/* 'out' allocated by encode function
 */
static void lz4pf_encode(void*_status,
			 const void*in, size_t in_size,
			 void**out, size_t* out_size)
{
  unsigned int s = LZ4_COMPRESSBOUND(in_size);
  void*p = padico_malloc(s);
  int rc = LZ4_compress((void*)in, p, in_size);
  if(rc > 0)
    {
      /* success */
      *out = p;
      *out_size = rc;
    }
  else
    {
      /* compression failed: don't compress */
      padico_fatal("LZ4: compression failed- rc=%d\n", rc);
      memcpy(p, in, in_size);
      *out = p;
      *out_size = in_size;
    }
}

/* 'out' allocated by caller
*/
static void lz4pf_decode(void*_status,
			  const void*in, size_t in_size,
			  void*out, size_t* out_size)
{
  unsigned int s = *out_size;
  int rc = LZ4_decompress_safe((void*)in, out, in_size, *out_size);
  if(rc == 0)
    {
      padico_fatal("LZ4: CRC error while uncompressing- in_size=%d; out_size=%d; rc=%d\n",
		   (int)in_size, (int)*out_size, rc);
    }
  *out_size = s;
}

static int lz4pf_init(void)
{
  padico_print("PacketFilter-LZ4: using LZ4 version %d.%d.%d\n", LZ4_VERSION_MAJOR, LZ4_VERSION_MINOR, LZ4_VERSION_RELEASE);

  puk_component_declare("PacketFilter-LZ4",
			puk_component_provides("PadicoComponent", "component", &lz4_component_driver),
			puk_component_provides("PacketFilter", "pf", &lz4pf_driver));
  return 0;
}
