/** @file
 * @brief Padico PadicoControl group control
 * @ingroup PadicoControl
 */

#include <Padico/Puk.h>
#include <Padico/Module.h>
#include <Padico/PadicoTM.h>
#include <Padico/Topology.h>

#include "PadicoControl-internals.h"

PADICO_MODULE_HOOK(PadicoControl);

/* ********************************************************* */

PUK_LIST_CREATE_TYPE(padico_group,
		     const char*group_id;
		     padico_na_node_vect_t nodes;
		     );
PUK_LIST_CREATE_FUNCS(padico_group);

/** a group request.
 * @note this is a struct only to allow a private definition of
 * the content of padico_group_req_t
 */
struct padico_group_req_s
{
  padico_reqs_vect_t reqs;
};

static padico_group_list_t na_groups;


/* ********************************************************* */

static void defgroup_end_handler(puk_parse_entity_t e)
{
  const int publish = puk_parse_getattr(e, "publish") != NULL;
  padico_group_t group = padico_group_init(puk_parse_getattr(e, "name"));

  puk_parse_entity_t*i;
  puk_vect_foreach(i, puk_parse_entity, puk_parse_get_contained(e))
    {
      padico_na_node_t node = puk_parse_get_content(*i);
      if(node == NULL)
	{
	  padico_warning("PadicoControl: bad node in group.\n");
	  puk_parse_set_rc(e, padico_rc_error("bad node in group"));
	  return;
	}
      padico_group_node_add(group, node);
    }
  if(publish)
    {
      padico_group_publish(group);
    }
  else
    {
      padico_group_store(group);
    }
}
static struct puk_tag_action_s defgroup_action =
{
  .xml_tag        = "PadicoControl:defgroup",
  .end_handler    = &defgroup_end_handler,
  .required_level = PUK_TRUST_CONTROL
};

/* ********************************************************* */

static void killgroup_end_handler(puk_parse_entity_t e)
{
  const char*groupid = puk_parse_getattr(e, "GroupID");
  assert(groupid != NULL);
  padico_group_t group = padico_group_lookup(groupid);
  assert(group != NULL);
  padico_group_kill(group, 0);
}

static struct puk_tag_action_s killgroup_action =
{
  .xml_tag        = "PadicoControl:killgroup",
  .start_handler  = NULL,
  .end_handler    = &killgroup_end_handler,
  .required_level = PUK_TRUST_CONTROL
};

/* ********************************************************* */

void padico_control_group_init(void)
{
  na_groups = padico_group_list_new();
  puk_xml_add_action(defgroup_action);
  puk_xml_add_action(killgroup_action);
}

/* *** Group ************************************************ */

padico_group_t padico_group_init(const char*id)
{
  padico_group_t g = padico_group_new();
  g->nodes    = padico_na_node_vect_new();
  g->group_id = padico_strdup(id);
  return g;
}

int padico_group_store(padico_group_t g)
{
  padico_trace("storing group %s nodes=%d\n",
	       g->group_id, padico_na_node_vect_size(g->nodes));
  padico_group_list_push_back(na_groups, g);
  return 0;
}

padico_group_t padico_group_lookup(const char*id)
{
  padico_group_itor_t i;
  for(i  = padico_group_list_begin(na_groups);
      i != padico_group_list_end(na_groups);
      i  = padico_group_list_next(i))
    {
      if((strcmp(i->group_id, id) == 0))
	{
	  return i;
	}
    }
  return NULL;
}

void padico_group_node_add(padico_group_t g, padico_na_node_t node)
{
  if(node != NULL)
    {
      padico_na_node_vect_push_back(g->nodes, node);
    }
  else
    {
      padico_warning("trying to add a NULL node to a group.\n");
    }
}
int padico_group_size(padico_group_t g)
{
  return padico_na_node_vect_size(g->nodes);
}

int padico_group_rank(padico_group_t g, padico_na_node_t node)
{
  int i;
  for(i = 0; i < padico_na_node_vect_size(g->nodes); i++)
    {
      if(padico_na_node_vect_at(g->nodes, i) == node)
	return i;
    }
  return -1;
}

const char*padico_group_name(padico_group_t group)
{
  return group->group_id;
}

padico_na_node_t padico_group_node(padico_group_t g, int rank)
{
  return padico_na_node_vect_at(g->nodes, rank);
}


void padico_group_destroy(padico_group_t g)
{
  padico_free((void*)g->group_id);
  padico_na_node_vect_delete(g->nodes);
  padico_group_list_erase(na_groups, g);
  padico_group_delete(g);
}


/** Publishes a group to its members. Should be called by the group leader.
 */
void padico_group_publish(padico_group_t group)
{ 
  padico_na_node_vect_itor_t nodes_itor;
  padico_string_t cmd = padico_string_new();
  padico_group_req_t greq;
  padico_group_store(group);
  if(padico_na_getlocalnode() == padico_na_node_vect_at(group->nodes, 0))
    {
      padico_string_printf(cmd, "<PadicoControl:defgroup name=\"%s\">\n", group->group_id);
      for(nodes_itor  = padico_na_node_vect_begin(group->nodes); 
	  nodes_itor != padico_na_node_vect_end(group->nodes);
	  nodes_itor  = padico_na_node_vect_next(nodes_itor))
	{
	  padico_string_catf(cmd, "  <Topology:node name=\"%s\"/>\n",
			     padico_na_node_getname(*nodes_itor));
	}
      padico_string_cat(cmd, "</PadicoControl:defgroup>\n");
      greq = padico_group_send_to_all(group, padico_string_get(cmd));
      padico_group_wait_all(group, greq);
    }
  padico_string_delete(cmd);
}

void padico_group_kill(padico_group_t group, int rc)
{
  padico_string_t cmd = padico_string_new();
  padico_string_printf(cmd, "<kill rc=\"%d\"/>", rc);
  int i;
  for(i = 0; i < padico_na_node_vect_size(group->nodes); i++)
    {
      const padico_na_node_t node = padico_na_node_vect_at(group->nodes, i);
      if(node != padico_na_getlocalnode())
	{
	  padico_control_send_oneway(node, padico_string_get(cmd));
	}
    }
  padico_string_delete(cmd);
}

padico_group_t padico_group_topology(padico_group_t g, const char*distrib)
{
  puk_hashtable_t table = puk_hashtable_new_ptr();
  padico_na_host_vect_t hosts = padico_na_host_vect_new();
  /* build hastable: host -> nodes vect; build hosts index *with order preserved* */
  padico_na_node_vect_itor_t i;
  puk_vect_foreach(i, padico_na_node, g->nodes)
    {
      padico_na_node_t n = *i;
      padico_na_host_t h = padico_na_node_gethost(n);
      padico_na_node_vect_t v = puk_hashtable_lookup(table, h);
      if(v == NULL)
	{
	  v = padico_na_node_vect_new();
	  puk_hashtable_insert(table, h, v);
	  padico_na_host_vect_push_back(hosts, h);
	}
      padico_na_node_vect_push_back(v, n);
    }
  /* compute matrix bounds */
  const int maxhosts = puk_hashtable_size(table);
  int maxnodes = 0;
  {
    puk_hashtable_enumerator_t e = puk_hashtable_enumerator_new(table);
    padico_na_node_vect_t v = puk_hashtable_enumerator_next_data(e);
    while(v != NULL)
      {
	const int size = padico_na_node_vect_size(v);
	if(size > maxnodes)
	  maxnodes = size;
	v = puk_hashtable_enumerator_next_data(e);
      }  
    puk_hashtable_enumerator_delete(e);
  }
  padico_na_node_t*matrix = calloc(maxhosts * maxnodes, sizeof(padico_na_node_t));
  padico_out(40, "maxhosts = %d; maxnodes = %d\n", maxhosts, maxnodes);
  /* store nodes at their new place in matrix */
  {
    padico_na_host_vect_itor_t h = NULL;
    puk_vect_foreach(h, padico_na_host, hosts)
      {
	padico_na_node_vect_t v = puk_hashtable_lookup(table, *h);
	puk_vect_foreach(i, padico_na_node, v)
	  {
	    const int nhost = padico_na_host_vect_rank(hosts, h);
	    const int nnode = padico_na_node_vect_rank(v, i);
	    padico_na_node_t n = *i;
	    if(strcmp(distrib, "cyclic") == 0)
	      {
		assert(matrix[nnode * maxhosts + nhost] == NULL);
		matrix[nnode * maxhosts + nhost] = n;
	      }
	    else if(strcmp(distrib, "block") == 0)
	      {
		assert(matrix[nhost * maxnodes + nnode] == NULL);
		matrix[nhost * maxnodes + nnode] = n;
	      }
	    else
	      {
		padico_fatal("NewMad_Launcher: unknown distribution %s\n", distrib);
	      }
	    *i = NULL;
	  }
      }
  }
  /* flatten matrix to eliminate holes, in case number of nodes is not multiple of number of hosts */
  padico_group_t newgroup = padico_group_init(distrib);
  int k;
  for(k = 0; k < maxhosts * maxnodes; k++)
    {
      padico_na_node_t n = matrix[k];
      if(n != NULL)
	{
	  padico_group_node_add(newgroup, n);
	  padico_out(40, "group: add node %s\n", padico_na_node_getname(n));
	}
    }
  /* cleanup */
  padico_na_node_vect_t v = NULL;
  do
    {
      puk_hashtable_getentry(table, NULL, v);
      if(v != NULL)
	padico_na_node_vect_delete(v);
    }
  while(v != NULL);
  puk_hashtable_delete(table);
  padico_na_host_vect_delete(hosts);
  free(matrix);
  return newgroup;
}


/* *** Group-control **************************************** */

/** Sends a command to each member of a group (except self)
 * @param group The destination group of nodes (local node should 
 *  be in the group, but won't receive the command)
 * @param cmd the command to send (NULL-terminated text, in XML)
 * @param greq a pointer to a grouped request descriptor
 */
padico_group_req_t padico_group_send_to_all(padico_group_t group, const char*cmd)
{
  padico_group_req_t greq = padico_malloc(sizeof(struct padico_group_req_s));
  int i;
  padico_out(40, "sending command to group %s\n%s", group->group_id, cmd);
  greq->reqs = padico_reqs_vect_new();
  for(i = 0; i < padico_na_node_vect_size(group->nodes); i++)
    {
      const padico_na_node_t node = padico_na_node_vect_at(group->nodes, i);
      padico_out(60, "  %s -> node %d/%d\n", cmd, i, padico_na_node_vect_size(group->nodes) );
      if(node != padico_na_getlocalnode())
	{
	  padico_req_t req = padico_control_send(node, cmd);
	  padico_reqs_vect_push_back(greq->reqs, req);
	}
    }
  return greq;
}

padico_rc_t padico_group_wait_all(padico_group_t group, padico_group_req_t greq)
{
  padico_rc_t rc = NULL;
  int i;
  padico_out(50, "waiting ACKs for group %s\n", group->group_id);
  for(i = 0; i < padico_reqs_vect_size(greq->reqs); i++)
    {
      padico_rc_t urc = NULL;
      padico_req_t req = padico_reqs_vect_at(greq->reqs, i); 
      padico_out(50, "waiting reply for grank=%d; req_id=%d\n",
		     i, req->id);
      urc = padico_tm_req_wait(req);
      rc = padico_rc_cat(rc, urc);
      if(urc)
	{
	  padico_rc_delete(urc);
	}
      padico_out(50, "got reply from node grank=%d\n", i);
    }
  padico_reqs_vect_delete(greq->reqs);
  padico_free(greq);
  padico_out(50, "request completed on group %s\n", group->group_id);
  return rc;
}

