/** @file Shm driver for NewMadeleine
 */

#include <nm_private.h>

#include <Padico/Module.h>
#include <Padico/Topology.h>

#include <limits.h>

#if defined(MARCEL)
#  warning "Shm: Marcel is enabled- not building module 'Shm'"
#else /* MARCEL */

PADICO_MODULE_DECLARE(NewMad_Driver_shm);
PADICO_MODULE_INIT(nm_shm_module_init);


#include "shm.h"

/** 'shm' driver per-instance data.
 */
struct nm_shm_drv
{
  struct padico_shm_s*shm;
  nm_gate_t gates[PADICO_SHM_NUMNODES];
  char*url;
  int nb_trks;
};

/** 'shm' per-gate data.
 */
struct nm_shm
{
  struct padico_shm_node_s*dest;
  int dest_rank;
};


/** shm NewMad Driver */
static const char*nm_shm_get_driver_url(nm_drv_t p_drv);
static int nm_shm_query(nm_drv_t p_drv, struct nm_driver_query_param *params, int nparam);
static int nm_shm_init(nm_drv_t p_drv, struct nm_trk_cap*trk_caps, int nb_trks);
static int nm_shm_exit(nm_drv_t p_drv);
static int nm_shm_connect(void*_status, nm_gate_t p_gate, nm_drv_t p_drv, nm_trk_id_t trk_id, const char*remote_url);
static int nm_shm_disconnect(void*_status, nm_gate_t p_gate, nm_drv_t p_drv, nm_trk_id_t trk_id);
static int nm_shm_send_iov(void*_status, struct nm_pkt_wrap_s *p_pw);
static int nm_shm_recv_iov(void*_status, struct nm_pkt_wrap_s *p_pw);
static int nm_shm_poll_send(void*_status, struct nm_pkt_wrap_s *p_pw);
static int nm_shm_poll_recv(void*_status, struct nm_pkt_wrap_s *p_pw);
static int nm_shm_cancel_recv_iov(void*_status, struct nm_pkt_wrap_s *p_pw);

static const struct nm_drv_iface_s nm_shm_driver =
  {
    .name               = "shm",
    
    .query              = &nm_shm_query,
    .init               = &nm_shm_init,
    .close              = &nm_shm_exit,

    .connect		= &nm_shm_connect,
    .disconnect         = &nm_shm_disconnect,

    .post_send_iov	= &nm_shm_send_iov,
    .poll_send_iov	= &nm_shm_poll_send,

    .post_recv_iov      = &nm_shm_recv_iov,
    .poll_recv_iov	= &nm_shm_poll_recv,
    .cancel_recv_iov    = &nm_shm_cancel_recv_iov,

    /* TODO: add poll_any callbacks  */
    .poll_send_any_iov  = NULL,
    .poll_recv_any_iov  = NULL,

    .get_driver_url     = &nm_shm_get_driver_url,

#ifdef PIOM_BLOCKING_CALLS
    .wait_send_iov	= NULL,
    .wait_recv_iov	= NULL,

    .wait_recv_any_iov  = NULL,
    .wait_send_any_iov  = NULL,
#endif
    
    .capabilities.is_exportable = 0,
    .capabilities.min_period    = 0,
    .capabilities.has_recv_any  = 1,
    .capabilities.max_unexpected = PADICO_SHM_SHORT_PAYLOAD
  };

/** 'PadicoComponent' facet for shm driver */
static void*nm_shm_instantiate(puk_instance_t, puk_context_t);
static void nm_shm_destroy(void*);

static const struct puk_component_driver_s nm_shm_component_driver =
  {
    .instantiate = &nm_shm_instantiate,
    .destroy     = &nm_shm_destroy
  };

/* ********************************************************* */

/** Component declaration */
int nm_shm_module_init(void)
{
  puk_component_declare("NewMad_Driver_shm",
			puk_component_provides("PadicoComponent", "component", &nm_shm_component_driver),
			puk_component_provides("NewMad_Driver", "driver", &nm_shm_driver));
  return 0;
}


/** Instanciate functions */
static void*nm_shm_instantiate(puk_instance_t instance, puk_context_t context)
{
  struct nm_shm*status = padico_malloc(sizeof(struct nm_shm));
  status->dest = NULL;
  status->dest_rank = -1;
  return status;
}

static void nm_shm_destroy(void*_status)
{
  padico_free(_status);
}

/** Url function */
static const char*nm_shm_get_driver_url(nm_drv_t p_drv)
{
  struct nm_shm_drv *p_shm_drv = p_drv->priv;
  return p_shm_drv->url;
}


static int nm_shm_query(nm_drv_t p_drv,
			struct nm_driver_query_param *params TBX_UNUSED,
			int nparam TBX_UNUSED)
{
  struct nm_shm_drv* p_shm_drv = TBX_MALLOC(sizeof(struct nm_shm_drv));
  memset(p_shm_drv, 0, sizeof(struct nm_shm_drv));
  p_drv->profile.latency = 200;
  p_drv->profile.bandwidth = 5000;
  p_drv->priv = p_shm_drv;
  return NM_ESUCCESS;
}

static int nm_shm_init(nm_drv_t p_drv, struct nm_trk_cap*trk_caps, int nb_trks)
{
  struct nm_shm_drv*p_shm_drv = p_drv->priv;
  p_shm_drv->shm = padico_shm_init("nmad");
  p_shm_drv->url = padico_strdup((char*)padico_na_node_getuuid(padico_na_getlocalnode()));

  /* open the requested number of tracks */
  int i;
  for(i = 0; i < nb_trks; i++)
    {
      /* track capabilities encoding */
      trk_caps[i].rq_type	= nm_trk_rq_dgram;
      trk_caps[i].iov_type	= nm_trk_iov_both_assym;
      trk_caps[i].max_pending_send_request  = 1;
      trk_caps[i].max_pending_recv_request  = 1;
      trk_caps[i].max_single_request_length = SSIZE_MAX;
      trk_caps[i].max_iovec_request_length  = 0;
#ifdef IOV_MAX
      trk_caps[i].max_iovec_size	    = IOV_MAX;
#else /* IOV_MAX */
      trk_caps[i].max_iovec_size	    = sysconf(_SC_IOV_MAX);
#endif /* IOV_MAX */
    }

  return NM_ESUCCESS;
}

static int nm_shm_exit(nm_drv_t p_drv)
{
  struct nm_shm_drv*p_shm_drv = p_drv->priv;
  assert(p_shm_drv != NULL);
  padico_shm_close(p_shm_drv->shm);
  padico_free(p_shm_drv->url);
  padico_free(p_shm_drv);
  p_drv->priv = NULL;
  return NM_ESUCCESS;
}

static int nm_shm_connect(void*_status, nm_gate_t p_gate, nm_drv_t p_drv, nm_trk_id_t trk_id, const char*remote_url)
{
  struct nm_shm*status = (struct nm_shm*)_status;
  struct nm_shm_drv*p_shm_drv = p_drv->priv;
  const char*remote_uuid = remote_url;
  padico_na_node_t node = padico_na_getnodebyuuid((padico_topo_uuid_t)remote_uuid);
  assert(node != NULL);
  if(status->dest == NULL)
    {
      struct padico_shm_node_s*dest = padico_shm_directory_node_lookup(p_shm_drv->shm, node);
      if(dest == NULL)
	{
	  fprintf(stderr, "nmad: FATAL- node %s not found in directory.\n", remote_uuid);
	  sleep(10000);
	  abort();
	}
      status->dest = dest;
      status->dest_rank = dest->rank;
      p_shm_drv->gates[dest->rank] = p_gate;
    }
  return NM_ESUCCESS;
}

static int nm_shm_disconnect(void*_status, nm_gate_t p_gate, nm_drv_t p_drv, nm_trk_id_t trk_id)
{
  struct nm_shm*status = (struct nm_shm*)_status;
  status->dest = NULL;
  return NM_ESUCCESS;
}

static int nm_shm_send_iov(void*_status, struct nm_pkt_wrap_s *p_pw)
{
  struct nm_shm*status = (struct nm_shm*)_status;
  struct nm_shm_drv*p_shm_drv = p_pw->p_drv->priv;
  const struct iovec*v = p_pw->v;
  const int n = p_pw->v_nb;
  if(p_pw->trk_id == NM_TRK_SMALL)
    {
      const int block_num = padico_shm_block_alloc(p_shm_drv->shm);
      void*base = padico_shm_short_get_ptr(p_shm_drv->shm, block_num);
      int i;
      for(i = 0; i < n; i++)
	{
	  memcpy(base, v[i].iov_base, v[i].iov_len);
	  base += v[i].iov_len;
	}
      assert(p_pw->length <= PADICO_SHM_SHORT_PAYLOAD);
      int rc = padico_shm_short_send_commit(p_shm_drv->shm, status->dest, block_num, p_pw->length, 1);
      if(rc)
	padico_fatal("NewMad_Shm: cannot send- receiver queue full.\n");
    }
  else
    {
      assert(n == 1);
      padico_shm_large_send(p_shm_drv->shm, v[0].iov_base, v[0].iov_len, status->dest);
    }
  return NM_ESUCCESS;
}

static int nm_shm_poll_send(void*_status, struct nm_pkt_wrap_s *p_pw)
{
  return NM_ESUCCESS;
}

static int nm_shm_recv_iov(void*_status, struct nm_pkt_wrap_s *p_pw)
{
  struct nm_shm*status = (struct nm_shm*)_status;
  struct nm_shm_drv*p_shm_drv = p_pw->p_drv->priv;
  if(p_pw->trk_id == NM_TRK_LARGE)
    {
      assert(p_pw->v_nb == 1);
      const int queue_num = padico_shm_large_recv_post(p_shm_drv->shm, status->dest_rank);
      p_pw->drv_priv = (void*)(intptr_t)queue_num;
      return -NM_EAGAIN;
    }
  else
    {
      return nm_shm_poll_recv(_status, p_pw);
    }
}


static int nm_shm_poll_recv(void*_status, struct nm_pkt_wrap_s *p_pw)
{
  struct nm_shm_drv*p_shm_drv = p_pw->p_drv->priv;
  struct iovec*v = p_pw->v;
  assert(p_pw->v_nb == 1);
  if(p_pw->trk_id == NM_TRK_SMALL)
    {
      const int block_num = padico_shm_short_recv_poll(p_shm_drv->shm, NULL);
      if(block_num != -1)
	{
#warning TODO- encapsulation
	  struct padico_shm_short_header_s*shm_header = padico_shm_block_get_ptr(p_shm_drv->shm, block_num);
	  const size_t size = shm_header->size;
	  const int from    = shm_header->from;
	  void*base = padico_shm_short_get_ptr(p_shm_drv->shm, block_num);
	  nm_gate_t p_gate = p_shm_drv->gates[from];
	  if(p_gate == NULL)
	    {
	      /* connection not established yet */
	      padico_shm_short_lfqueue_enqueue(&p_shm_drv->shm->self->mailbox.shared_queue, block_num);
	      return -NM_EAGAIN;
	    }
	  assert(p_pw->p_gate == NULL || p_pw->p_gate == p_gate);
	  p_pw->p_gate = p_gate;
	  assert(size <= v[0].iov_len);
	  v[0].iov_len = size;
	  memcpy(v[0].iov_base, base, size);
	  padico_shm_block_free(p_shm_drv->shm, block_num);
	  return NM_ESUCCESS;
	}
      else
	{
	  return -NM_EAGAIN;
	}
    }
  else
    {
      const int queue_num = (int)(intptr_t)p_pw->drv_priv;
      int rc = padico_shm_large_recv_poll(p_shm_drv->shm, v[0].iov_base, v[0].iov_len, queue_num);
      if(rc == 0)
	return NM_ESUCCESS;
      else
	return -NM_EAGAIN;
    }
  return NM_ESUCCESS;
}

static int nm_shm_cancel_recv_iov(void*_status, struct nm_pkt_wrap_s *p_pw)
{
  if(p_pw->trk_id == NM_TRK_SMALL)
    {
      p_pw->drv_priv = NULL;
      return NM_ESUCCESS;
    }
  else
    return -NM_ENOTIMPL;
}

#endif /* MARCEL */
