/** @file
 * @brief Grid topology directory
 * @ingroup Topology
 */

#include <Padico/Puk.h>
#include <Padico/Module.h>
#include <Padico/PM2.h>
#include <Padico/PadicoTM.h>

#include "Topology-internals.h"

#include <string.h>
#include <unistd.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <resolv.h>

#ifdef HAVE_IFADDRS_H
#include <ifaddrs.h>
#include <netpacket/packet.h>
#endif
#include <sys/ioctl.h>

#ifdef HAVE_LINUX_NETLINK_H
#include <linux/netlink.h>
#endif
#ifdef HAVE_LINUX_RTNETLINK_H
#include <linux/rtnetlink.h>
#endif
#ifdef HAVE_LINUX_SOCKIOS_H
#include <linux/sockios.h>
#endif
#ifdef HAVE_LINUX_ETHTOOL_H
#include <linux/ethtool.h>
#endif
#ifdef HAVE_LINUX_IF_H
#include <linux/if.h>
#else
#include <net/if.h>
#endif /* HAVE_LINUX_IF_H */


PADICO_MODULE_DECLARE(Topology)
PADICO_MODULE_INIT(padico_topology_init);

__PUK_SYM_INTERNAL struct padico_topology_s na_topology;


static void topology_populate(void);
static void topo_inet_addr_detect(padico_na_host_t localhost);
static void topo_inet_route_detect(padico_na_host_t localhost);
static void topo_inet_service_detect(padico_na_host_t localhost);
static void topo_device_detect(void);
static const char*topo_addr_visibility_name(padico_topo_addr_visibility_t v);
static const char*topo_hwaddr_type_name(int type);


/* ********************************************************* */

/* Helper for addresses in bindings */

const struct topo_address_s*topo_binding_getaddr(const struct topo_binding_s*_binding)
{
  struct topo_binding_s*const binding = (struct topo_binding_s*)_binding; /* wild cast to deal with iterators, but we don't write binding, really */
  topo_property_vect_itor_t p;
  puk_vect_foreach(p, topo_property, &binding->props)
    {
      const struct topo_property_s*prop = *p;
      if(prop->kind == TOPO_PROPERTY_ADDR)
	{
	  return &prop->content.ADDR.addr;
	}
    }
  return NULL;
}

/* hashing functions for addresses */

static int topo_address_eq(const void*_addr1, const void*_addr2)
{
  const struct topo_address_s*addr1 = _addr1;
  const struct topo_address_s*addr2 = _addr2;
  return (addr1->size == addr2->size) &&
         (memcmp(&addr1->_addr_storage[0], &addr2->_addr_storage[0], addr1->size) == 0);
}
static uint32_t topo_address_hash(const void*_addr)
{
  const struct topo_address_s*addr = _addr;
  return puk_hash_oneatatime((const void*)&addr->_addr_storage[0], addr->size);
}

/* hashing functions for UUID */

static int topo_uuid_eq(const void*uuid1, const void*uuid2)
{
  return (memcmp(uuid1, uuid2, PADICO_TOPO_UUID_SIZE) == 0);
}
static uint32_t topo_uuid_hash(const void*uuid)
{
  return puk_hash_oneatatime(uuid, PADICO_TOPO_UUID_SIZE);
}

/* ********************************************************* */
/* ** hwaddr topology modification */

padico_topo_hwaddr_t padico_topo_hwaddr_create(const char*name, int index, int type, size_t len, const void*addr)
{
  struct topo_hwaddr_s*hwaddr = padico_malloc(sizeof(struct topo_hwaddr_s) + len);
  hwaddr->type = type;
  hwaddr->index = index;
  hwaddr->len = len;
  memcpy(&hwaddr->_hwaddr_storage, addr, len);
  assert(strlen(name) < TOPO_HWADDR_NAMSIZ);
  strncpy(hwaddr->name, name, TOPO_HWADDR_NAMSIZ);
  hwaddr->name[TOPO_HWADDR_NAMSIZ - 1] = 0;
  return hwaddr;
}

void padico_topo_hwaddr_add(padico_na_host_t host, padico_topo_hwaddr_t*_hwaddr)
{
  padico_topo_hwaddr_t hwaddr = *_hwaddr;
  padico_topo_hwaddr_t byname = topo_hwaddr_getbyname(host, hwaddr->name);
  if(byname == NULL)
    {
      /* store new hwaddr */
      topo_hwaddr_vect_push_back(&host->hwaddrs, hwaddr);
    }
  else if((byname->len == hwaddr->len) && (memcmp(byname->_hwaddr_storage, hwaddr->_hwaddr_storage, byname->len) == 0))
    {
      /* duplicate detected (same name, same address, index not compared since it may be invalid at this stage) */
      padico_free((void*)hwaddr);
      hwaddr = NULL;
      *_hwaddr = byname;
    }
  else
    {
      padico_fatal("Topology: conflicting hwaddr with same name=%s; different address detected.\n", hwaddr->name);
    }
}

/* ********************************************************* */
/* ** binding topology modification */

void padico_topo_binding_add_addr(padico_topo_binding_t binding, const void*addr)
{
  padico_topo_network_t network = binding->network;
  const size_t addr_size = network->addr_size;
  const int family = network->family;
  struct topo_property_s*prop = padico_malloc(sizeof(struct topo_property_s) + addr_size);
  prop->kind = TOPO_PROPERTY_ADDR;
  prop->content.ADDR.addr.family = family;
  prop->content.ADDR.addr.size = addr_size;
  memcpy(&prop->content.ADDR.addr._addr_storage[0], addr, addr_size);
  topo_property_vect_push_back(&binding->props, prop);
  /*  puk_hashtable_insert(binding->network->bindings_by_addr, &prop->content.ADDR._addr_storage, binding); */
}

void padico_topo_binding_add_hwaddr(padico_topo_binding_t binding, padico_topo_hwaddr_t hwaddr)
{
  struct topo_property_s*prop = padico_malloc(sizeof(struct topo_property_s));
  prop->kind = TOPO_PROPERTY_HWADDR;
  prop->content.HWADDR.hwaddr = hwaddr;
  topo_property_vect_push_back(&binding->props, prop);
}

void padico_topo_binding_add_gw_addr(padico_topo_binding_t binding, struct in_addr*inaddr)
{
  topo_property_vect_itor_t p;
  struct topo_property_s*prop = NULL;
  puk_vect_foreach(p, topo_property, &binding->props)
    {
      prop = *p;
      if(prop->kind == TOPO_PROPERTY_GW_ADDR &&
	 memcmp(&prop->content.GW_ADDR.addr._addr_storage[0], inaddr,  prop->content.GW_ADDR.addr.size) == 0)
	return;
    }
  prop = padico_malloc(sizeof(struct topo_property_s) + sizeof(struct in_addr));
  prop->kind = TOPO_PROPERTY_GW_ADDR;
  prop->content.GW_ADDR.addr.family = AF_INET;
  prop->content.GW_ADDR.addr.size = sizeof(struct in_addr);
  memcpy(&prop->content.GW_ADDR.addr._addr_storage[0], inaddr, sizeof(struct in_addr));
  topo_property_vect_push_back(&binding->props, prop);
}

void padico_topo_binding_add_gw_hwaddr(padico_topo_binding_t binding, padico_topo_hwaddr_t hwaddr)
{
  struct topo_property_s*prop = padico_malloc(sizeof(struct topo_property_s));
  prop->kind = TOPO_PROPERTY_GW_HWADDR;
  prop->content.GW_HWADDR.hwaddr = hwaddr;
  topo_property_vect_push_back(&binding->props, prop);
}

void padico_topo_binding_add_hostname(padico_topo_binding_t binding, const char*name)
{
  struct topo_property_s*prop = padico_malloc(sizeof(struct topo_property_s));
  prop->kind = TOPO_PROPERTY_HOSTNAME;
  prop->content.HOSTNAME.name = padico_strdup(name);
  topo_property_vect_push_back(&binding->props, prop);
}

void padico_topo_binding_add_route(padico_topo_binding_t binding, padico_topo_network_t dest, const void*gwaddr)
{
  padico_topo_network_t network = binding->network;
  const size_t addr_size = network->addr_size;
  const int family = network->family;
  struct topo_property_s*prop = padico_malloc(sizeof(struct topo_property_s) + addr_size);
  prop->kind = TOPO_PROPERTY_ROUTE;
  prop->content.ROUTE.dest = dest;
  prop->content.ROUTE.gw.family = family;
  prop->content.ROUTE.gw.size = addr_size;
  memcpy(&prop->content.ROUTE.gw._addr_storage[0], gwaddr, addr_size);
  topo_property_vect_push_back(&binding->props, prop);
}

void padico_topo_binding_add_filter(padico_topo_binding_t binding, padico_topo_filter_policy_t policy, uint16_t port, int protocol)
{
  assert(policy > _TOPO_FILTER_NONE && policy < _TOPO_FILTER_MAX);
  struct topo_property_s*prop = padico_malloc(sizeof(struct topo_property_s));
  prop->kind = TOPO_PROPERTY_FILTER;
  prop->content.FILTER.filter.policy = policy;
  prop->content.FILTER.filter.port = port;
  prop->content.FILTER.filter.protocol = protocol;
  topo_property_vect_push_back(&binding->props, prop);
}

void padico_topo_binding_add_perf(padico_topo_binding_t binding, int latency, int bandwidth)
{
  struct topo_property_s*prop = padico_malloc(sizeof(struct topo_property_s));
  prop->kind = TOPO_PROPERTY_PERF;
  prop->content.PERF.perf.lat = latency;
  prop->content.PERF.perf.bw = bandwidth;
  topo_property_vect_push_back(&binding->props, prop);
}

/* ********************************************************* */

int padico_topology_init(void)
{

  sethostent(1);

  na_topology.networks    = padico_na_network_vect_new();
  na_topology.node_table  = puk_hashtable_new(&topo_uuid_hash, &topo_uuid_eq);
  na_topology.nodesbyname = puk_hashtable_new_string();
  na_topology.hostsbyname = puk_hashtable_new_string();
  na_topology.inet        = padico_topo_network_create("inet", PADICO_NA_SCOPE_HOST, sizeof(struct in_addr), AF_INET);
  na_topology.inet6       = padico_topo_network_create("inet6", PADICO_NA_SCOPE_HOST, sizeof(struct in6_addr), AF_INET6);
  topology_parser_init();
  topology_populate();
  return 0;
}

/** frontend to gethostbyaddr */
static const struct hostent*topo_gethostbyaddr(const void*addr, socklen_t len, int family)
{
  static int disable_ipv6 = -1;
  if(disable_ipv6 < 0)
    {
      disable_ipv6 = (padico_getenv("PADICO_DISABLE_IPV6") != NULL);
      if(disable_ipv6)
	{
	  padico_warning("Topology: IPv6 disabled by user request.\n");
	}
    }
  if(padico_getenv("PADICO_DISABLE_DNS") != NULL)
    {
      return NULL;
    }
  char s_addr[256] = { '\0' };
  inet_ntop(family, addr, s_addr, 256);
  if(disable_ipv6 && family == AF_INET6)
    {
      return NULL;
    }
  const struct hostent*he = gethostbyaddr(addr, len, family);
  return he;
}

/** frontend to gethostbyname2 to manage TRY_AGAIN error */
static const struct hostent*topo_gethostbyname2(const char*name, int family)
{
  static int disable_ipv6 = -1;
  if(disable_ipv6 < 0)
    {
      disable_ipv6 = (padico_getenv("PADICO_DISABLE_IPV6") != NULL);
      if(disable_ipv6)
	{
	  padico_warning("Topology: IPv6 disabled by user request.\n");
	}
    }
  h_errno = 0;
  if(disable_ipv6 && family == AF_INET6)
    {
      return NULL;
    }
  const struct hostent*he = gethostbyname2(name, family);
  if(!he)
    {
      int count = 5;
      while(he == NULL && h_errno == TRY_AGAIN && count > 0)
	{
	  h_errno = 0;
	  he = gethostbyname2(name, family);
	  count--;
	}
    }
  if(!he)
    {
      /* some resolver are picky with short names without final dot */
      if(strchr(name, '.') == NULL)
	{
	  char*tmp_name = padico_malloc(strlen(name) + 2);
	  strcpy(tmp_name, name);
	  strcat(tmp_name, ".");
	  he = gethostbyname2(tmp_name, family);
	  padico_free(tmp_name);
	}
    }
  return he;
}

static void topo_detect_link_perf(padico_topo_binding_t b, const padico_topo_hwaddr_t hwaddr)
{
#if defined(HAVE_LINUX_SOCKIOS_H) && defined(HAVE_LINUX_ETHTOOL_H) && defined(HAVE_LINUX_IF_H)
  int sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_IP);
   if(sock < 0)
     return;
   struct ifreq ifr;
   struct ethtool_cmd edata;
   int rc;
   strncpy(ifr.ifr_name, hwaddr->name, sizeof(ifr.ifr_name));
   ifr.ifr_data = &edata;
   edata.cmd = ETHTOOL_GSET;
   rc = ioctl(sock, SIOCETHTOOL, &ifr);
   if (rc < 0)
     goto close;
   padico_topo_binding_add_perf(b, 300000, edata.speed*1000/8);
 close:
   close(sock);
#endif
}

/** detect network interfaces using Linux getifaddrs, if available
 */
static int topo_inet_addr_detect_getifaddrs(padico_na_host_t localhost)
{
#ifdef HAVE_IFADDRS_H
  struct ifaddrs*ifa_list = NULL;
  int rc = getifaddrs(&ifa_list);
  if(rc == 0)
    {
      const struct ifaddrs*i;
      for(i = ifa_list; i != NULL; i = i->ifa_next)
	{
	  const int is_loopback = !!(i->ifa_flags & IFF_LOOPBACK);
	  const int is_up = !!(i->ifa_flags & IFF_UP);
	  if (i->ifa_addr && i->ifa_addr->sa_family == AF_INET)
	    {
	      /* ** detect IPv4 address */
	      const struct sockaddr_in*inaddr = (const struct sockaddr_in*)i->ifa_addr;
	      padico_topo_addr_visibility_t v = padico_topo_inet_visibility(inaddr->sin_addr.s_addr);
	      const uint32_t netmask = ntohl(((struct sockaddr_in*)i->ifa_netmask)->sin_addr.s_addr);
	      int prefix_len = 0;
	      while((prefix_len < 32) && ((netmask & (0x01 << (31 - prefix_len))) != 0))
		{
		  prefix_len ++;
		}
	      padico_print("Topology: address  %8s; inet=%15s/%2d (%s)\n",
			   i->ifa_name, inet_ntoa(inaddr->sin_addr),
			   prefix_len, topo_addr_visibility_name(v));
	      
	      if(is_up)
		{
		  padico_topo_hwaddr_t hwaddr = topo_hwaddr_getbyname(localhost, i->ifa_name);
		  const struct hostent*he = topo_gethostbyaddr(&inaddr->sin_addr, sizeof(struct sockaddr_in), AF_INET);
		  if(he)
		    padico_topo_host_add_name(localhost, he->h_name, -1);
		  switch(v)
		    {
		    case TOPO_ADDR_VISIBILITY_PUBLIC:
		    case TOPO_ADDR_VISIBILITY_LINKLOCAL:
		      {
			padico_topo_binding_t binding = 
			  padico_na_host_bind(localhost, na_topology.inet, &inaddr->sin_addr);
			if(hwaddr)
			  padico_topo_binding_add_hwaddr(binding, hwaddr);
			if(he)
			  padico_topo_binding_add_hostname(binding, he->h_name);
		      }
		      break;
		    case TOPO_ADDR_VISIBILITY_PRIVATE:
		      {
			padico_string_t network_name = padico_string_new();
			const struct in_addr network_addr = { .s_addr = inaddr->sin_addr.s_addr & htonl(netmask) };
			padico_string_printf(network_name, "inet-[%s/%d]", inet_ntoa(network_addr), prefix_len);
			padico_topo_network_t network = padico_na_network_getbyid(padico_string_get(network_name));
			if(network == NULL)
			  {
			    network = padico_topo_network_create(padico_string_get(network_name),
								 PADICO_NA_SCOPE_HOST, sizeof(struct in_addr), AF_INET);
			  }
			padico_topo_binding_t priv_binding = padico_na_host_bind(localhost, network, &inaddr->sin_addr);
			if(hwaddr)
			  {
			    padico_topo_binding_add_hwaddr(priv_binding, hwaddr);
			    topo_detect_link_perf(priv_binding, hwaddr);
			  }
			if(he)
			  padico_topo_binding_add_hostname(priv_binding, he->h_name);
			padico_string_delete(network_name);
		      }
		      break;
		    case TOPO_ADDR_VISIBILITY_LOOPBACK:
		      {
			padico_string_t network_name = padico_string_new();
			padico_string_printf(network_name, "inet-[loopback/%s]", localhost->hostname);
			padico_topo_network_t network = padico_na_network_getbyid(padico_string_get(network_name));
			if(network == NULL)
			  {
			    network = padico_topo_network_create(padico_string_get(network_name),
								 PADICO_NA_SCOPE_HOST, sizeof(struct in_addr), AF_INET);
			  }
			padico_na_host_bind(localhost, network, &inaddr->sin_addr);
			padico_string_delete(network_name);
		      }
		      break;
		    case TOPO_ADDR_VISIBILITY_NONE:
		    default:
		      padico_warning("Topology: not storing address in binding.\n");
		    }		   
		}
	    }
	  else if(i->ifa_addr && i->ifa_addr->sa_family == AF_INET6)
	    {
	      /* ** detect IPv6 address */
	      const struct sockaddr_in6*in6addr = (const struct sockaddr_in6*)i->ifa_addr;
	      padico_topo_addr_visibility_t v = padico_topo_inet6_visibility(&in6addr->sin6_addr);
	      char addr[INET6_ADDRSTRLEN];
	      inet_ntop(AF_INET6, &in6addr->sin6_addr, addr, INET6_ADDRSTRLEN);
	      padico_print("Topology: address  %8s; inet6=%46s (%s)\n", i->ifa_name, addr, topo_addr_visibility_name(v));
	      if(is_up && !is_loopback && v == TOPO_ADDR_VISIBILITY_PUBLIC)
		{
		  padico_topo_binding_t binding = 
		    padico_na_host_bind(localhost, na_topology.inet6, &in6addr->sin6_addr);
		  padico_topo_hwaddr_t hwaddr = topo_hwaddr_getbyname(localhost, i->ifa_name);
		  if(hwaddr)
		    padico_topo_binding_add_hwaddr(binding, hwaddr);
		}
	    }
	  else if(i->ifa_addr && i->ifa_addr->sa_family == AF_PACKET)
	    {
	      /* ** detect hardware address */
	      const struct sockaddr_ll*lladdr = (const struct sockaddr_ll*)i->ifa_addr;
	      padico_string_t s_addr = padico_string_new();
	      int j;
	      for(j = 0; j < lladdr->sll_halen; j++)
		{
		  padico_string_catf(s_addr, (j==0)?"%02X":":%02X", lladdr->sll_addr[j]);
		}
	      padico_print("Topology: hardware %8s; hwaddr=%s; index=%d (%s)\n",
			   i->ifa_name, padico_string_get(s_addr),
			   lladdr->sll_ifindex, topo_hwaddr_type_name(lladdr->sll_hatype));
	      padico_string_delete(s_addr);
	      if(lladdr->sll_hatype != ARPHRD_LOOPBACK)
		{
		  padico_topo_hwaddr_t hwaddr = 
		    padico_topo_hwaddr_create(i->ifa_name, lladdr->sll_ifindex, lladdr->sll_hatype, lladdr->sll_halen, lladdr->sll_addr);
		  padico_topo_hwaddr_add(localhost, &hwaddr);
		}
	    }
	  else if(i->ifa_addr)
	    {
	      padico_print("Topology: unknown address family (%d)\n", i->ifa_addr->sa_family);
	    }
	}
    } 
  else
    {
      padico_warning("Topology: error while listing network interfaces (getifaddrs: %s).\n", strerror(errno));
      return -1;
    }
  freeifaddrs(ifa_list);
  return 0;
#else /* HAVE_IFADDRS_H */
  return -1;
#endif /* HAVE_IFADDRS_H */
}

/** detect network interfaces using netdevice ioctl
 */
static void topo_inet_addr_detect_siocgifconf(padico_na_host_t localhost)
{
  const size_t max_ifindex = 32;
  struct ifreq ifreqs[max_ifindex];
  struct ifconf ifc =
    {
      .ifc_len = max_ifindex * sizeof(struct ifreq), 
      .ifc_buf = (void*)&ifreqs
    };
  memset(&ifreqs[0], 0, max_ifindex * sizeof(struct ifreq));
  int sockfd = PUK_ABI_FSYS_WRAP(socket)(AF_INET, SOCK_DGRAM, 0); /* socket for netdevice ioctl() */
  if(sockfd < 0)
    {
      padico_warning("Topology: error while enumerating addresses.\n");
      return;
    }
  int rc = PUK_ABI_FSYS_WRAP(ioctl)(sockfd, SIOCGIFCONF, &ifc);
  /* warning- SIOCGIFCONF only returns IPv4 addresses */
  if(rc)
    padico_fatal("Topology: error %s while enumerating addresses.\n",
		 strerror(__puk_abi_wrap__errno));
  int i;
  for(i = 0; i < ifc.ifc_len / sizeof(struct ifreq); i++)
    {
      struct ifreq*ifr = &ifreqs[i];
      padico_print("Topology: IF addr family=%d; if=%s\n",
		   ifr->ifr_addr.sa_family, ifr->ifr_name);
      PUK_ABI_FSYS_WRAP(ioctl)(sockfd, SIOCGIFHWADDR, ifr);
      padico_print("Topology: IF hw addr family=%d; if=%s\n",
		   ifr->ifr_addr.sa_family, ifr->ifr_name);
    }
  PUK_ABI_FSYS_WRAP(close)(sockfd);
}

/** detect network interfaces using netlink socket
 */
static int topo_inet_addr_detect_netlink(padico_na_host_t localhost)
{
#if defined(HAVE_LINUX_NETLINK_H) && defined(HAVE_LINUX_RTNETLINK_H)
  /* create netlink socket */
  int fd = PUK_ABI_FSYS_WRAP(socket)(AF_NETLINK, SOCK_RAW, NETLINK_ROUTE);
  if(fd < 0)
    {
      padico_warning("Topology: cannot open netlink socket (socket: %s).\n", strerror(errno));
      return -1;
    }
  struct sockaddr_nl addr = {
    .nl_family = AF_NETLINK,
    .nl_pad    = 0,
    .nl_pid    = getpid(),
    .nl_groups = 0
  };
  int rc = PUK_ABI_FSYS_WRAP(bind)(fd, (struct sockaddr*)&addr, sizeof(addr));
  if(rc < 0)
    {
      padico_warning("Topology: cannot bind netlink socket.\n");
      return -1;
    }
  /* netlink request */
  struct 
  {
    struct nlmsghdr nl; /**< netlink header */
    struct ifaddrmsg ifa;
    char            buf[8192];
  } req;
  memset(&req, 0, sizeof(req));
  req.nl.nlmsg_len   = NLMSG_LENGTH(sizeof(struct ifaddrmsg));
  req.nl.nlmsg_type  = RTM_GETADDR;
  req.nl.nlmsg_flags = NLM_F_REQUEST | NLM_F_DUMP;
  req.nl.nlmsg_seq   = 0;
  req.nl.nlmsg_pid   = getpid();
  req.ifa.ifa_family = AF_INET;
  struct iovec iov = { .iov_base = (void*)&req.nl, .iov_len = req.nl.nlmsg_len };
  /* send request */
  struct sockaddr_nl peer = {
    .nl_family = AF_NETLINK,
    .nl_pad    = 0,
    .nl_pid    = 0, /* 0 = kernel */
    .nl_groups = 0
  };
  struct msghdr msg = {
    .msg_iov     = &iov,
    .msg_iovlen  = 1,
    .msg_name    = (void*)&peer,
    .msg_namelen = sizeof(peer)
  };
  rc = PUK_ABI_FSYS_WRAP(sendmsg)(fd, &msg, 0);
  if(rc < 0)
    padico_fatal("Topology: error while sending netlink request.\n");
  /* receive reply */
  char buf[8192];
  memset(buf, 0, sizeof(buf));
  char*p = buf;
  int nll = 0;
  while(1) 
    {
      rc = PUK_ABI_FSYS_WRAP(recv)(fd, p, sizeof(buf) - nll, 0);
      struct nlmsghdr*nlp = (struct nlmsghdr*)p;
      if(nlp->nlmsg_type == NLMSG_DONE)
	break;
      p += rc;
      nll += rc;
    }
  /* parse reply */
  const struct nlmsghdr*nlp = (struct nlmsghdr*)buf;
  while(NLMSG_OK(nlp, nll))
    {
      const struct ifaddrmsg*ifa = (struct ifaddrmsg*)NLMSG_DATA(nlp);
      const struct rtattr*rtap = (struct rtattr*)IFA_RTA(ifa);
      int rtl = IFA_PAYLOAD(nlp);
      while(RTA_OK(rtap, rtl))
	{
	  switch(rtap->rta_type)
	    {
	    case IFA_LOCAL:
	      {
		if(ifa->ifa_family == AF_INET)
		  {
		    const struct in_addr*inaddr = (const struct in_addr*)RTA_DATA(rtap);
		    const int prefix_len = ifa->ifa_prefixlen;
		    const int ifindex = ifa->ifa_index;
		    padico_topo_addr_visibility_t v = padico_topo_inet_visibility(inaddr->s_addr);
		    padico_print("Topology: netlink   index=%d; inet=%15s/%2d (%s)\n",
				 ifindex, inet_ntoa(*inaddr), prefix_len, topo_addr_visibility_name(v));
		    switch(v)
		      {
		      case TOPO_ADDR_VISIBILITY_PUBLIC:
		      case TOPO_ADDR_VISIBILITY_LINKLOCAL:
			{
			  padico_na_host_bind(localhost, na_topology.inet, inaddr);
			}
			break;
		      case TOPO_ADDR_VISIBILITY_PRIVATE:
			{
			  padico_warning("Topology: TODO- private address\n");
			  padico_na_host_bind(localhost, na_topology.inet, inaddr);
			  /*
			  padico_string_t network_name = padico_string_new();
			  const struct in_addr network_addr = { .s_addr = inaddr->s_addr & htonl(netmask) };
			  padico_string_printf(network_name, "inet-[%s/%d]", inet_ntoa(network_addr), prefix_len);
			  padico_topo_network_t network = padico_na_network_getbyid(padico_string_get(network_name));
			  if(network == NULL)
			    {
			      network = padico_topo_network_create(padico_string_get(network_name),
								   PADICO_NA_SCOPE_HOST, sizeof(struct in_addr), AF_INET);
			    }
			  padico_topo_binding_t priv_binding = padico_na_host_bind(localhost, network, &inaddr->sin_addr);
			  padico_string_delete(network_name);
			  */
			}
			break;
		      case TOPO_ADDR_VISIBILITY_LOOPBACK:
			padico_warning("Topology: TODO- loopback\n");
			break;
		      case TOPO_ADDR_VISIBILITY_NONE:
		      default:
			padico_warning("Topology: not storing address in binding.\n");
			
		      }
		  }
	      }
	      break;
	    default:
	      break;
	    }
	  rtap = RTA_NEXT(rtap,rtl);
	}
      nlp = NLMSG_NEXT(nlp, nll);
    }


  PUK_ABI_FSYS_WRAP(close)(fd);
  return 0;
#else
  return -1;
#endif /* HAVE_LINUX_NETLINK_H && HAVE_LINUX_RTNETLINK_H */
}

static void topo_inet_addr_detect(padico_na_host_t localhost)
{
  /* list network interfaces */
#ifdef MCKERNEL
  topo_inet_addr_detect_netlink(localhost);
#else /* MCKERNEL */
  int rc = topo_inet_addr_detect_getifaddrs(localhost);
  if(rc != 0)
    {
      rc = topo_inet_addr_detect_netlink(localhost);
      if(rc != 0)
	{
	  topo_inet_addr_detect_siocgifconf(localhost);
	}
    }
#endif /* MCKERNEL */
}

static void topo_inet_route_detect(padico_na_host_t localhost)
{
#if defined(HAVE_LINUX_NETLINK_H) && defined(HAVE_LINUX_RTNETLINK_H)
  /* create rtnetlink socket */
  int fd = PUK_ABI_FSYS_WRAP(socket)(AF_NETLINK, SOCK_RAW, NETLINK_ROUTE);
  if(fd < 0)
    {
      padico_warning("Topology: cannot open rtnetlink socket (socket: %s).\n", strerror(errno));
      return;
    }
  struct sockaddr_nl addr = {
    .nl_family = AF_NETLINK,
    .nl_pad    = 0,
    .nl_pid    = getpid(),
    .nl_groups = 0
  };
  int rc = PUK_ABI_FSYS_WRAP(bind)(fd, (struct sockaddr*)&addr, sizeof(addr));
  if(rc < 0)
    padico_fatal("Topology: cannot bind rtnetlink socket.\n");
  /* rtnetlink request */
  struct 
  {
    struct nlmsghdr nl; /**< netlink header */
    struct rtmsg    rt; /**< rtnetlink request */
    char            buf[8192];
  } req;
  memset(&req, 0, sizeof(req));
  req.nl.nlmsg_len   = NLMSG_LENGTH(sizeof(struct rtmsg));
  req.nl.nlmsg_type  = RTM_GETROUTE;
  req.nl.nlmsg_flags = NLM_F_REQUEST | NLM_F_DUMP;
  req.nl.nlmsg_seq   = 0;
  req.nl.nlmsg_pid   = getpid();
  req.rt.rtm_family  = AF_INET;
  req.rt.rtm_table   = RT_TABLE_MAIN;
  req.rt.rtm_type    = RTN_UNICAST; /* gateway */
  struct iovec iov = { .iov_base = (void*)&req.nl, .iov_len = req.nl.nlmsg_len };
  /* send request */
  struct sockaddr_nl peer = {
    .nl_family = AF_NETLINK,
    .nl_pad    = 0,
    .nl_pid    = 0, /* 0 = kernel */
    .nl_groups = 0
  };
  struct msghdr msg = {
    .msg_iov     = &iov,
    .msg_iovlen  = 1,
    .msg_name    = (void*)&peer,
    .msg_namelen = sizeof(peer)
  };
  rc = PUK_ABI_FSYS_WRAP(sendmsg)(fd, &msg, 0);
  if(rc < 0)
    padico_fatal("Topology: error while sending rtnetlink request.\n");
  /* receive reply */
  char buf[8192];
  memset(buf, 0, sizeof(buf));
  char*p = buf;
  int nll = 0;
  while(1) 
    {
      rc = PUK_ABI_FSYS_WRAP(recv)(fd, p, sizeof(buf) - nll, 0);
      struct nlmsghdr*nlp = (struct nlmsghdr*)p;
      if(nlp->nlmsg_type == NLMSG_DONE)
	break;
      p += rc;
      nll += rc;
    }
  /* parse reply */
  const struct nlmsghdr*nlp = (struct nlmsghdr*)buf;
  while(NLMSG_OK(nlp, nll))
    {
      const struct rtmsg*rtp = (struct rtmsg*)NLMSG_DATA(nlp);
      if(rtp->rtm_table == RT_TABLE_MAIN) 
	{
	  int ifn = -1;
	  struct in_addr gw = { .s_addr = INADDR_NONE };
	  struct in_addr dest = { .s_addr = INADDR_NONE };
	  const struct rtattr*rtap = (struct rtattr*)RTM_RTA(rtp);
	  int rtl = RTM_PAYLOAD(nlp);
	  while(RTA_OK(rtap, rtl))
	    {
	      switch(rtap->rta_type)
		{
		case RTA_DST:     /* destination IPv4 address */
		  dest = *(struct in_addr*)RTA_DATA(rtap);
		  break;
		case RTA_GATEWAY: /* gateway IPv4 address */
		  gw = *(struct in_addr*)RTA_DATA(rtap);
		  break;
		case RTA_OIF:     /* outbound interface number */
		  ifn = *(int*)RTA_DATA(rtap);
		default:
		  break;
		}
	      rtap = RTA_NEXT(rtap,rtl);
	    }
	  if(ifn != -1 && gw.s_addr != INADDR_NONE)
	    {
	      padico_topo_hwaddr_t hwaddr = topo_hwaddr_getbyindex(localhost, ifn);
	      char*dest_s = padico_strdup(inet_ntoa(dest));
	      padico_print("Topology: route    %8s; gw = %15s; dest = %15s/%d\n", hwaddr->name, inet_ntoa(gw), dest_s, rtp->rtm_dst_len);
	      struct ifreq ifr;
	      struct arpreq arpreq;
	      int sockfd = PUK_ABI_FSYS_WRAP(socket)(AF_INET, SOCK_DGRAM, 0);
	      ifr.ifr_addr.sa_family = AF_INET;
	      strncpy(ifr.ifr_name, hwaddr->name, TOPO_HWADDR_NAMSIZ);
	      PUK_ABI_FSYS_WRAP(ioctl)(sockfd, SIOCGIFADDR, &ifr);
	      memset(&arpreq, 0, sizeof(struct arpreq));
	      struct sockaddr_in*gwaddr = (struct sockaddr_in*)&arpreq.arp_pa;
	      memcpy(&gwaddr->sin_addr, &gw, sizeof(struct in_addr));
	      gwaddr->sin_family = AF_INET;
	      strcpy(arpreq.arp_dev, hwaddr->name);
	      padico_topo_hwaddr_t gw_hwaddr = NULL;
	      if(PUK_ABI_FSYS_WRAP(ioctl)(sockfd, SIOCGARP, &arpreq) == 0)
		{
		  gw_hwaddr = padico_topo_hwaddr_create("", 0, ARPHRD_ETHER, 6, &arpreq.arp_ha.sa_data[0]);
		}
	      PUK_ABI_FSYS_WRAP(close)(sockfd);
	      struct in_addr local = ((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr;
	      padico_topo_addr_visibility_t local_v = padico_topo_inet_visibility(local.s_addr);
	      padico_topo_addr_visibility_t dest_v = padico_topo_inet_visibility(dest.s_addr);
	      if(local_v == TOPO_ADDR_VISIBILITY_PRIVATE)
		{
		  /* find binding for given hwaddr */
		  topo_binding_vect_itor_t b;
		  puk_vect_foreach(b, topo_binding, &localhost->bindings)
		    {
		      padico_topo_binding_t binding = *b;
		      topo_property_vect_itor_t p;
		      puk_vect_foreach(p, topo_property, &binding->props)
			{
			  if((*p)->kind == TOPO_PROPERTY_HWADDR && (*p)->content.HWADDR.hwaddr == hwaddr)
			    {
			      padico_topo_binding_add_gw_addr(binding, &gw);
			      if(gw_hwaddr)
				padico_topo_binding_add_gw_hwaddr(binding, gw_hwaddr);
			      if(dest.s_addr == INADDR_NONE)
				{
				  padico_topo_network_t network =  binding->network;
				  if(!network->uplink  && network->family == AF_INET)
				    {
				      padico_topo_binding_t uplink = topo_binding_create(na_topology.inet);
				      uplink->entity.network = network;
				      uplink->entity.kind = TOPO_ENTITY_NETWORK;
				      network->uplink = uplink;
				      padico_topo_binding_add_filter(uplink, TOPO_FILTER_BLOCK_IN, PORT_ANY, IPPROTO_TCP);
				      padico_topo_binding_add_filter(uplink, TOPO_FILTER_BLOCK_IN, PORT_ANY, IPPROTO_UDP);
				    }
				}
			      if(dest.s_addr != INADDR_NONE && dest_v == TOPO_ADDR_VISIBILITY_PRIVATE)
				{
				  padico_string_t network_name = padico_string_new();
				  padico_string_printf(network_name, "inet-[%s/%d]", dest_s, rtp->rtm_dst_len);
				  padico_topo_network_t network = padico_na_network_getbyid(padico_string_get(network_name));
				  if(!network)
				    {
				      network = padico_topo_network_create(padico_string_get(network_name),
									   PADICO_NA_SCOPE_HOST, sizeof(struct in_addr), AF_INET);
				      padico_topo_binding_add_route(binding, network, &gw);
				    }
				}
			      break; /* props have been modified- iterator is now invalid */
			    }
			}
		    }
		}
	      padico_free(dest_s);
	    }
	}
      nlp = NLMSG_NEXT(nlp, nll);
    }
  PUK_ABI_FSYS_WRAP(close)(fd);
#endif
}

static void topo_inet_service_detect(padico_na_host_t localhost)
{
  unsigned long txq, rxq, retrnsmt, inode, tm;
  int sl, local_port, remote_port, state, tr, uid, timeout;
  char line[2048], local_addr[128], remote_addr[128], end[1024];
  size_t line_len = 0, rc = 0;
  line_len = sizeof(line);
  FILE *tcp_file = fopen("/proc/net/tcp", "r");
  setservent(1);
  if(tcp_file)
    {
      fgets(line, line_len, tcp_file);
      do
	{
	  if(!fgets(line, line_len, tcp_file))
	    break;
	  rc = sscanf(line, "%d: %[0-9A-Fa-f]:%X %64[0-9A-Fa-f]:%X %X %lX:%lX %X:%lX %lX %d %d %ld %512s\n",
		      &sl, local_addr, &local_port, remote_addr, &remote_port, &state, &txq, &rxq, &tr, &tm, &retrnsmt, &uid, &timeout, &inode, end);
	  if(rc <= 0)
	    break;
	  struct in_addr local_inaddr;
	  sscanf(local_addr, "%X", &(local_inaddr.s_addr));
	  local_inaddr.s_addr = htonl(local_inaddr.s_addr);
	  if(state == TCP_LISTEN && local_inaddr.s_addr != INADDR_LOOPBACK && local_port < 1024)
	    {
	      struct servent*servent = getservbyport(htons(local_port), "tcp");
	      if(servent)
		padico_topo_host_add_service(localhost, servent->s_name, ntohs(servent->s_port));
	    }
	}
      while(!feof(tcp_file));
      fclose(tcp_file);
    }
  endservent();
}

/* if host have special devices, try load driver */
static void topo_device_detect(void)
{
  const char*autodetect = padico_getattr("PADICO_AUTODETECT");
  if(!autodetect || (autodetect && strcmp(autodetect, "no")))
    {
      struct stat s;
      int ret;
      padico_rc_t rc = NULL;
      ret = stat("/dev/infiniband", &s);
      if(!ret && S_ISDIR(s.st_mode))
	{
	  puk_mod_t mod = NULL;
	  rc = padico_puk_mod_resolve(&mod, "InfinibandVerbs");
	  if(mod)
	    {
	      rc = padico_puk_mod_load(mod);
	    }
	}
      /* TODO- add Myrinet detection */
      if(rc)
	padico_rc_delete(rc);
    }
}

padico_topo_addr_visibility_t padico_topo_inet_visibility(in_addr_t _inaddr)
{
  in_addr_t inaddr = ntohl(_inaddr);
  padico_topo_addr_visibility_t v = TOPO_ADDR_VISIBILITY_NONE;
  if((inaddr & 0xff000000) == 0x7f000000)           /* 127.0.0.x/8 */
    v = TOPO_ADDR_VISIBILITY_LOOPBACK;
  else if( ((inaddr & 0xff000000) == 0x0a000000) || /* 10.x.x.x/8 */
	   ((inaddr & 0xfff00000) == 0xac100000) || /* 172.16.x.x/12 */
	   ((inaddr & 0xffff0000) == 0xc0a80000) )  /* 192.168.x.x/16 */
    v = TOPO_ADDR_VISIBILITY_PRIVATE;
  else if( ((inaddr & 0xffff0000) == 0xa9fe0000) || /* 169.254.x.x/16*/
	   ((inaddr & 0xffffff00) == 0xc0000200) )  /* 192.0.2.x/24 */
    v = TOPO_ADDR_VISIBILITY_LINKLOCAL;
  else
    v = TOPO_ADDR_VISIBILITY_PUBLIC;
  return v;
}

padico_topo_addr_visibility_t padico_topo_inet6_visibility(const struct in6_addr*in6addr)
{
  if(memcmp(in6addr, &in6addr_loopback, sizeof(struct in6_addr)) == 0)
    return TOPO_ADDR_VISIBILITY_LOOPBACK;
  else if((ntohs(in6addr->s6_addr16[0]) & 0xffc0) == 0xfe80) /* fe80::/10 */
    return TOPO_ADDR_VISIBILITY_LINKLOCAL;
  else if((ntohs(in6addr->s6_addr16[0]) & 0xfe00) == 0xfc00) /* fc00::/7 */
    return TOPO_ADDR_VISIBILITY_PRIVATE;
  else
    return TOPO_ADDR_VISIBILITY_PUBLIC;
}

/** inet address visibility pretty print */
static const char*topo_addr_visibility_name(padico_topo_addr_visibility_t v)
{
  switch(v)
    {
    case TOPO_ADDR_VISIBILITY_NONE:
      return "none";
    case TOPO_ADDR_VISIBILITY_LOOPBACK:
      return "loopback";
    case TOPO_ADDR_VISIBILITY_LINKLOCAL:
      return "link-local";
    case TOPO_ADDR_VISIBILITY_PRIVATE:
      return "private";
    case TOPO_ADDR_VISIBILITY_PUBLIC:
      return "public";
    default:
      return "unknown";
    }
}

/** MAC address type pretty print */
static const char*topo_hwaddr_type_name(int type)
{
  switch(type)
    {
    case  ARPHRD_ETHER:
    case ARPHRD_EETHER:
      return "ethernet";
    case ARPHRD_LOOPBACK:
      return "loopback";
#ifdef ARPHRD_ATM
    case ARPHRD_ATM:
      return "atm";
#endif
#ifdef ARPHRD_IEEE1394
    case ARPHRD_IEEE1394:
      return "ieee1394";
#endif
#ifdef ARPHRD_INFINIBAND
    case ARPHRD_INFINIBAND:
      return "infiniband";
#endif
#ifdef ARPHRD_PPP
    case ARPHRD_PPP:
	return "ppp";
#endif
#ifdef ARPHRD_IRDA
    case  ARPHRD_IRDA:
      return "IrDA";
#endif
#ifdef ARPHRD_IEEE80211
    case ARPHRD_IEEE80211:
      return "ieee802.11";
#endif
    default:
      return "unknown";
    }
}

static void topology_populate(void)
{
  const char*attr_uuid = padico_getattr("PADICO_NODE_UUID"); 
  char*uuid = NULL;
  if(attr_uuid != NULL)
    {
      uuid = padico_strdup(attr_uuid);
    }
  else
    {
      uuid = padico_malloc(sizeof(struct topo_uuid_s));
      srandom(getpid());
      int i;
      for(i = 0 ; i< PADICO_TOPO_UUID_SIZE; i++)
	{
	  static const char digits[] = "0123456789ABCDEF\0";
	  uuid[i] = digits[random() % 16];
	}
      uuid[PADICO_TOPO_UUID_SIZE] = 0;
    }

  /* create localhost */
  const int local_hostname_size = 257;
  char local_hostname[local_hostname_size];
  gethostname(local_hostname, local_hostname_size);
  const struct hostent*he = topo_gethostbyname2(local_hostname, AF_INET);
  padico_na_host_t localhost = NULL;
  if(he == NULL)
    {
      padico_warning("Topology: local hostname = %s is not resolvable (%s).\n", local_hostname, hstrerror(h_errno));
      localhost = padico_topo_host_create(local_hostname);
    }
  else
    {
      localhost = padico_topo_host_create(he->h_name);
      padico_topo_host_add_name(localhost, local_hostname, -1);
      const int uid = getuid();
      struct passwd*login = getpwuid(uid);
      if(login)
	padico_topo_host_add_property(localhost, "login", login->pw_name);
    }

  /* create node name */
  padico_string_t nodename = padico_string_new();
  const int pid = getpid();
  const char*boot_id   = padico_getattr("PADICO_BOOT_ID");
  const char*boot_rank = padico_getattr("PADICO_BOOT_RANK");
  int rank = boot_rank ? atoi(boot_rank) : 0;
  padico_string_printf(nodename, "%s#%d/%s", localhost->hostname, pid, boot_id);
  padico_print("Topology: this node is- %s\n", padico_string_get(nodename));

  /* create node  */
  na_topology.local_node = 
    topo_node_create(padico_string_get(nodename), localhost, uuid, boot_id);
  padico_string_delete(nodename);
  padico_free(uuid);

  /* loopback and session networks */
  padico_string_t s_session = padico_string_new();
  padico_string_printf(s_session, "Session-[%s]", boot_id);
  padico_topo_network_t session = padico_topo_network_create(padico_string_get(s_session), PADICO_NA_SCOPE_NODE, sizeof(int), AF_UNSPEC);
  padico_na_node_bind(na_topology.local_node, session, &rank);
  padico_string_delete(s_session);
  padico_string_t s_hloop = padico_string_new();
  padico_string_printf(s_hloop, "LoopbackHost-[%s]", localhost->hostname);
  padico_topo_network_t hloopback = padico_topo_network_create(padico_string_get(s_hloop), PADICO_NA_SCOPE_NODE, sizeof(int), AF_UNSPEC);
  padico_na_node_bind(na_topology.local_node, hloopback, &pid);
  padico_string_delete(s_hloop);
  padico_string_t s_nloop = padico_string_new();
  padico_string_printf(s_nloop, "LoopbackNode-[%s]", padico_na_node_getname(na_topology.local_node));
  padico_topo_network_t nloopback = padico_topo_network_create(padico_string_get(s_nloop), PADICO_NA_SCOPE_NODE, sizeof(int), AF_UNSPEC);
  padico_na_node_bind(na_topology.local_node, nloopback, &pid);
  padico_string_delete(s_nloop);

  /* detecte inet features */
  padico_print("Topology: this host is %s- detecting networks...\n", localhost->hostname);
  topo_inet_addr_detect(localhost);
  topo_inet_route_detect(localhost);
  topo_inet_service_detect(localhost);

  /* detecte special device */
  topo_device_detect();
}

/* ********************************************************* */

/** Create a new 'host' record from a given canonical name
 * @note this function takes ownership of memory pointed by 'fqdn' (supposed to be malloced)
 */
padico_na_host_t padico_topo_host_create(const char*hostname)
{
  padico_na_host_t host = padico_malloc(sizeof(struct topo_host_s));
  padico_na_node_vect_init(&host->nodes);
  topo_binding_vect_init(&host->bindings);
  topo_hwaddr_vect_init(&host->hwaddrs);
  topo_hostname_vect_init(&host->hnames);
  topo_property_vect_init(&host->props);
  host->hostname = padico_strdup(hostname);
  padico_topo_host_add_name(host, hostname, -1);
  padico_print("Topology: creating host %s.\n", host->hostname);
  return host;
}

padico_na_host_t topo_host_lookup(const char*hostname)
{
  padico_na_host_t host = puk_hashtable_lookup(na_topology.hostsbyname, hostname);
  if(!host)
    {
      if(strcmp(hostname, "localhost") == 0 ||
	 strcmp(hostname, "ip6-localhost") == 0 ||
	 strcmp(hostname, "ip6-loopback") == 0)
	{
	  host = padico_na_getlocalhost();
	}
    }
  return host;
}

void padico_topo_host_add_name(padico_na_host_t host, const char*_name, int resolvable)
{
  if(strcmp(_name, "localhost") == 0)
    {
      return;
    }
  topo_hostname_vect_itor_t i;
  puk_vect_foreach(i, topo_hostname, &host->hnames)
    {
      if(strcmp((*i)->name, _name) == 0)
	return;
    }
  if(resolvable == -1)
    {
      if(padico_getenv("PADICO_DISABLE_DNS") != NULL)
	{
	  padico_warning("Topology: DNS query disabled by user request.\n");
	  resolvable = 0;
	}
      else
	{
	  unsigned char reply[1024];
	  int len = 1024;
	  int rc = res_init();
	  if(rc != 0)
	    padico_warning("Topology: cannot initialize resovler.\n");
	  rc = res_query(_name, ns_c_in, ns_t_a, reply, len);
	  if(rc == -1)
	    resolvable = 0;
	  else
	    resolvable = 1;
	}
    }
  struct topo_hostname_s*hname = padico_malloc(sizeof(struct topo_hostname_s));
  hname->name = padico_strdup(_name);
  hname->resolvable = resolvable;
  topo_hostname_vect_push_back(&host->hnames, hname);
  puk_hashtable_insert(na_topology.hostsbyname, hname->name, host);
} 

void padico_topo_host_add_service(padico_na_host_t host, const char*service_name, uint16_t service_port)
{
  struct topo_property_s*prop;
  topo_property_vect_itor_t p;
  size_t name_len = strlen(service_name);
  puk_vect_foreach(p, topo_property, &host->props)
    {
      prop = *p;
      if(prop->kind == TOPO_PROPERTY_SERVICE)
	{
	  padico_topo_service_t service = &prop->content.SERVICE.service;
	  if(!strcmp(service->name, service_name)
	     && service_port == service->port)
	    return;
	}
    }
  prop = padico_malloc(sizeof(struct topo_property_s) + name_len + 1);
  prop->kind = TOPO_PROPERTY_SERVICE;
  padico_topo_service_t service = &prop->content.SERVICE.service;
  memcpy(&service->name[0], service_name, name_len);
  service->name[name_len] = 0;
  service->port = service_port;
  topo_property_vect_push_back(&host->props, prop);
}

void padico_topo_host_add_property(padico_na_host_t host, const char*label, const char*value)
{
  const char*find = padico_na_host_getproperty(host, label);
  if(!find || (find && strcmp(find, value)))
    {
      struct topo_property_s*prop = padico_malloc(sizeof(struct topo_property_s));
      prop->kind = TOPO_PROPERTY_MISC;
      prop->content.MISC.label = padico_strdup(label);
      prop->content.MISC.value = padico_strdup(value);
      topo_property_vect_push_back(&host->props, prop);
    }
}

/* creates a binding between a host and a network */
padico_topo_binding_t padico_na_host_bind(padico_na_host_t host, padico_topo_network_t network, const void*address)
{
  padico_topo_binding_t binding = topo_binding_create(network);
  padico_topo_binding_add_addr(binding, address);
  topo_binding_bind_host(&binding, host);
  return binding;
}

/* creates a binding between a node and a network */
padico_topo_binding_t padico_na_node_bind(padico_na_node_t node, padico_topo_network_t network, const void*address)
{
  padico_topo_binding_t binding = topo_binding_create(network);
  padico_topo_binding_add_addr(binding, address);
  topo_binding_bind_node(&binding, node);
  return binding;
}

void topo_binding_bind_host(padico_topo_binding_t*_binding, padico_na_host_t host)
{
  padico_topo_binding_t binding = *_binding;
  assert(binding != NULL);
  /* resolve hwaddr stored by name */
  topo_property_vect_itor_t p;
  puk_vect_foreach(p, topo_property, &binding->props)
    {
      if(((*p)->kind == TOPO_PROPERTY_HWADDR) && ((*p)->content.HWADDR.hwaddr == NULL))
	{
	  padico_topo_hwaddr_t hwaddr = topo_hwaddr_getbyname(host, (*p)->content.HWADDR.ifname);
	  (*p)->content.HWADDR.hwaddr = hwaddr;
	}
    }
  topo_binding_vect_itor_t b;
  puk_vect_foreach(b, topo_binding, &host->bindings)
    {
      if((*b)->network == binding->network)
	{
	  const struct topo_address_s*addr1 = topo_binding_getaddr(binding);
	  const struct topo_address_s*addr2 = topo_binding_getaddr(*b);
	  if(topo_address_eq(addr1, addr2))
	    {
	      topo_binding_merge(*b, binding);
	      *_binding = *b;
	      return; /* avoid double-binding */
	    }
	}
    }
  assert(binding->network->scope == PADICO_NA_SCOPE_HOST);
  binding->entity.host = host;
  binding->entity.kind = TOPO_ENTITY_HOST;
  topo_binding_vect_push_back(&host->bindings, binding);
}

void topo_binding_bind_node(padico_topo_binding_t*_binding, padico_na_node_t node)
{
  padico_topo_binding_t binding = *_binding;
  assert(binding != NULL);
  topo_binding_vect_itor_t b;
  puk_vect_foreach(b, topo_binding, &node->bindings)
    {
      if((*b)->network == binding->network)
	{
	  const struct topo_address_s*addr1 = topo_binding_getaddr(binding);
	  const struct topo_address_s*addr2 = topo_binding_getaddr(*b);
	  if(topo_address_eq(addr1, addr2))
	    {
	      topo_binding_merge(*b, binding);
	      *_binding = *b;
	      return; /* avoid double-binding */
	    }
	}
    }
  assert(binding->network->scope == PADICO_NA_SCOPE_NODE);
  binding->entity.node = node;
  binding->entity.kind = TOPO_ENTITY_NONE;
  topo_binding_vect_push_back(&node->bindings, binding);
}


/** Create a new node, either locally or received as XML description
 */
padico_na_node_t topo_node_create(const char*name,
				  padico_na_host_t host,
				  const char*uuid,
				  const char*boot_id)
{
  padico_na_node_t node = padico_malloc(sizeof(struct topo_node_s));;
  padico_print("Topology: new node- [%s] %s\n", uuid, name);
  /* create a node */
  topo_binding_vect_init(&node->bindings);
  padico_na_node_vect_push_back(&host->nodes, node);
  node->host               = host;
  node->canonical_nodename = padico_strdup(name);
  node->boot_id            = padico_strdup(boot_id);
  assert(uuid != NULL);
  assert(strlen(uuid) == PADICO_TOPO_UUID_SIZE);
  memcpy(node->uuid.bytes, uuid, PADICO_TOPO_UUID_SIZE);
  node->uuid.zero = 0;
  puk_hashtable_insert(na_topology.node_table, node->uuid.bytes, node);
  puk_hashtable_insert(na_topology.nodesbyname, node->canonical_nodename, node);
  return node;
}

padico_topo_network_t padico_topo_network_create(const char*id, padico_na_binding_scope_t scope, size_t addr_size, int family)
{
  padico_topo_network_t network = padico_na_network_getbyid(id);
  if(network != NULL)
    {
      padico_warning("Topology: duplicate declaration for network %s\n", id);
      return network;
    }
  network = padico_malloc(sizeof(struct topo_network_s));
  assert(scope == PADICO_NA_SCOPE_NODE || scope == PADICO_NA_SCOPE_HOST);
  topo_binding_vect_init(&network->bindings);
  network->bindings_by_addr = puk_hashtable_new(&topo_address_hash, &topo_address_eq);
  network->network_id = padico_strdup(id);
  network->scope = scope;
  network->addr_size = addr_size;
  network->family = family;
  network->uplink = NULL;
  network->perf.lat = 0;
  network->perf.bw = 0;
  padico_na_network_vect_push_back(na_topology.networks, network);
  return network;
}

/* *** Topology inspectors ********************************* */


/* *** local info */

padico_na_node_t padico_na_getlocalnode(void)
{
  return na_topology.local_node;
}

/* *** node inspectors */

padico_topo_address_t padico_na_node_getaddress(padico_na_node_t node, padico_topo_network_t network)
{
  topo_binding_vect_itor_t i;
  puk_vect_foreach(i, topo_binding, &node->bindings)
    {
      const struct topo_binding_s*const binding = *i;
      if(binding->network == network && network->scope == PADICO_NA_SCOPE_NODE)
	{
	  return topo_binding_getaddr(binding);
	}
    }
  return NULL;
}

padico_topo_uuid_t padico_na_node_getuuid(padico_na_node_t node)
{
  return &node->uuid;
}

const char*padico_na_node_getname(padico_na_node_t node)
{
  return node->canonical_nodename;
}

const char*padico_na_node_getsession(padico_na_node_t node)
{
  return node->boot_id;
}

padico_na_host_t padico_na_node_gethost(padico_na_node_t node)
{
  return node->host;
}

void padico_na_node_getnetworks(padico_na_node_t node, padico_na_network_vect_t networks)
{
  topo_binding_vect_itor_t b;
  puk_vect_foreach(b, topo_binding, &node->bindings)
    {
      padico_na_network_vect_push_back(networks, (*b)->network);
    }
}


/* *** host inspector */

const char*padico_na_host_getname(padico_na_host_t host)
{
  return host->hostname;
}

const char*padico_na_host_getname2(padico_na_host_t host, padico_topo_network_t network)
{
  topo_binding_vect_itor_t i;
  puk_vect_foreach(i, topo_binding, &host->bindings)
    {
      if((*i)->network == network)
	{
	  topo_property_vect_itor_t p;
	  puk_vect_foreach(p, topo_property, &(*i)->props)
	    {
	      struct topo_property_s*prop = *p;
	      if(prop->kind == TOPO_PROPERTY_HOSTNAME)
		{
		  return prop->content.HOSTNAME.name;
		}
	    }
	}
    }
  topo_hostname_vect_itor_t h;
  puk_vect_foreach(h, topo_hostname, &host->hnames)
    {
      if((*h)->resolvable)
	return (*h)->name;
    }
  const struct in_addr*addr = padico_na_host_getaddr2(host, network);
  return inet_ntoa(*addr);
}

const struct in_addr*padico_na_host_getaddr(padico_na_host_t host)
{
  const struct in_addr*inaddr = NULL;
  padico_topo_addr_visibility_t max_v = TOPO_ADDR_VISIBILITY_NONE;
  topo_binding_vect_itor_t i;
  /* find a binding for network 'inet' */
  puk_vect_foreach(i, topo_binding, &host->bindings)
    {
      if((*i)->network == na_topology.inet)
	{
	  const struct in_addr*_inaddr = topo_address_as_inaddr(topo_binding_getaddr(*i));
	  padico_topo_addr_visibility_t v = padico_topo_inet_visibility(_inaddr->s_addr);
	  if(v > max_v)
	    {
	      inaddr = _inaddr;
	      max_v = v;
	    }
	}
    }
  /* in case of failure, find a binding for any AF_INET network */
  if(!inaddr)
    {
      puk_vect_foreach(i, topo_binding, &host->bindings)
	{
	  if((*i)->network->family == AF_INET)
	    {
	      const struct in_addr*_inaddr = topo_address_as_inaddr(topo_binding_getaddr(*i));
	      padico_topo_addr_visibility_t v = padico_topo_inet_visibility(_inaddr->s_addr);
	      if(v > max_v)
		{
		  inaddr = _inaddr;
		  max_v = v;
		}
	    }
	}
    }
  if(!inaddr)
    {
      padico_warning("Topology: host %s has no inet address.\n", host->hostname);
    }
  return inaddr;
}

const struct in_addr*padico_na_host_getaddr2(padico_na_host_t host, padico_topo_network_t network)
{
  const struct in_addr*inaddr = NULL;
  topo_binding_vect_itor_t i;
  puk_vect_foreach(i, topo_binding, &host->bindings)
    {
      if((*i)->network == network && 
	 (*i)->network->family == AF_INET)
	{
	  inaddr = topo_address_as_inaddr(topo_binding_getaddr(*i));
	}
    }
  return inaddr;
}

padico_na_node_vect_t padico_na_host_getnodes(padico_na_host_t host)
{
  return &host->nodes;
}

void padico_na_host_getnetworks(padico_na_host_t host, padico_na_network_vect_t networks)
{
  topo_binding_vect_itor_t b;
  puk_vect_foreach(b, topo_binding, &host->bindings)
    {
      padico_topo_network_t network = (*b)->network;
      padico_na_network_vect_push_back(networks, network);
      while(network->uplink != NULL)
      	{
      	  network = network->uplink->network;
      	  padico_na_network_vect_push_back(networks, network);
      	}
    }
}

static padico_topo_filter_policy_t padico_topo_host_policy(topo_net_entity_t entity, padico_topo_network_t network, uint16_t port, int protocol)
{
  padico_topo_filter_policy_t policy = 0;
  if((network->family == AF_INET || network->family == AF_INET6) && protocol)
    {
      /* find binding between host and network */
      padico_topo_binding_t binding = NULL;
      if(entity.kind == TOPO_ENTITY_HOST)
	{
	  padico_na_host_t host = entity.host;
	  topo_binding_vect_itor_t b;
	  puk_vect_foreach(b, topo_binding, &host->bindings)
	    {
	      padico_topo_network_t network2 = (*b)->network;
	      if(network2 == network)
		{
		  binding = *b;
		  break;
		}
	      while(network2->uplink != NULL)
		{
		  if(network2->uplink->network == network)
		    {
		      binding = network2->uplink;
		      break;
		    }
		  else
		    {
		      network2 = network2->uplink->network;
		    }
		}
	    }
	}
      else if(entity.kind == TOPO_ENTITY_NETWORK)
	{
	  if(entity.network->uplink != NULL && entity.network->uplink->network == network)
	    binding = entity.network->uplink;
	}
      /* compute policy from binding */
      if(binding)
	{
	  topo_property_vect_itor_t p;
	  puk_vect_foreach(p, topo_property, &binding->props)
	    {
	      const struct topo_property_s*prop = *p;
	      if((prop->kind == TOPO_PROPERTY_FILTER)
		 &&
		 (prop->content.FILTER.filter.protocol == protocol)
		 &&
		 (prop->content.FILTER.filter.port == port ||
		  prop->content.FILTER.filter.port == PORT_ANY))
		{
		  if( ((prop->content.FILTER.filter.policy & TOPO_FILTER_MASK_OUT) && ((policy & TOPO_FILTER_MASK_OUT) == 0))
		      ||
		      ((prop->content.FILTER.filter.policy & TOPO_FILTER_MASK_IN) && ((policy & TOPO_FILTER_MASK_IN) == 0)))
		    {
		      policy |= prop->content.FILTER.filter.policy;
		    }
		}
	    }
	  /* apply default policy */
	  if(!(policy & TOPO_FILTER_MASK_IN))
	    policy |= TOPO_FILTER_ALLOW_IN;
	  if(!(policy & TOPO_FILTER_MASK_OUT))
	    policy |= TOPO_FILTER_ALLOW_OUT;
	}
    }
  else
    {
      /* default policy for non-IP network */
      policy |= TOPO_FILTER_ALLOW_IN;
      policy |= TOPO_FILTER_ALLOW_OUT;
    }
  return policy;
}


int padico_na_host_can_input(padico_na_host_t host, padico_topo_network_t network, uint16_t port, int protocol)
{
  struct topo_net_entity_s e;
  e.kind = TOPO_ENTITY_HOST;
  e.host = host;
  return !!(padico_topo_host_policy(e, network, port, protocol) & TOPO_FILTER_ALLOW_IN);
}

int padico_na_host_can_output(padico_na_host_t host, padico_topo_network_t network, uint16_t port, int protocol)
{
  struct topo_net_entity_s e;
  e.kind = TOPO_ENTITY_HOST;
  e.host = host;
  return !!(padico_topo_host_policy(e, network, port, protocol) & TOPO_FILTER_ALLOW_OUT);
}

int padico_topo_network_can_input(padico_topo_network_t subnet, padico_topo_network_t network, uint16_t port, int protocol)
{
  struct topo_net_entity_s e;
  e.kind = TOPO_ENTITY_NETWORK;
  e.network = subnet;
  return !!(padico_topo_host_policy(e, network, port, protocol) & TOPO_FILTER_ALLOW_IN);
}

int padico_topo_network_can_output(padico_topo_network_t subnet, padico_topo_network_t network, uint16_t port, int protocol)
{
  struct topo_net_entity_s e;
  e.kind = TOPO_ENTITY_NETWORK;
  e.network = subnet;
  return !!(padico_topo_host_policy(e, network, port, protocol) & TOPO_FILTER_ALLOW_OUT);
}

int padico_na_host_getservice(padico_na_host_t host, char*service_name)
{
  struct topo_property_s*prop;
  topo_property_vect_itor_t p;
  puk_vect_foreach(p, topo_property, &host->props)
    {
      prop = *p;
      if(prop->kind == TOPO_PROPERTY_SERVICE)
	{
	  padico_topo_service_t service = &prop->content.SERVICE.service;
	  if(!strcmp(service->name, service_name))
	    return service->port;
	}
    }
  return 0;
}

const char*padico_na_host_getproperty(padico_na_host_t host, const char*label)
{
  struct topo_property_s*prop;
  topo_property_vect_itor_t p;
  puk_vect_foreach(p, topo_property, &host->props)
    {
      prop = *p;
      if(prop->kind == TOPO_PROPERTY_MISC)
	{
	  if(!strcmp((*p)->content.MISC.label, label))
	    return (*p)->content.MISC.value;
	}
    }
  return NULL;
}

padico_na_host_vect_t padico_na_gethosts()
{
  padico_na_host_vect_t hosts = padico_na_host_vect_new();
  puk_hashtable_enumerator_t e;
  char*name; 
  e = puk_hashtable_enumerator_new(na_topology.hostsbyname);
  padico_na_host_t host;
  puk_hashtable_enumerator_next2(e, &name, &host);
  while(host)
    {
      if(strcmp(name, host->hostname) == 0)
	{
	  padico_na_host_vect_push_back(hosts, host);
	}
      puk_hashtable_enumerator_next2(e, &name, &host);
    }
  puk_hashtable_enumerator_delete(e);
  return hosts;
}

/* *** network inspector */

const char*padico_na_network_getname(padico_topo_network_t network)
{
  return network->network_id;
}
int padico_na_network_size(padico_topo_network_t network)
{
  return topo_binding_vect_size(&network->bindings);
}
static inline int padico_na_network_inet_info(const padico_topo_network_t network, struct in_addr*addr, int*cidr)
{
  char addr_s[16];
  memset(addr_s, 0, 16);
  if(sscanf(network->network_id, "inet-[%[^/]/%d]", addr_s, cidr) != 2)
    return 0;
  return inet_pton(AF_INET, addr_s, addr);
}
static inline int na_bindings_vect_contains_network(struct topo_binding_vect_s*bindings, padico_topo_network_t network)
{
  topo_binding_vect_itor_t b;
  puk_vect_foreach(b, topo_binding, bindings)
    {
      if((*b)->network == network)
	return 1;
    }
  return 0;
}
int padico_na_network_contains_node(padico_topo_network_t network, padico_na_node_t node)
{
  return na_bindings_vect_contains_network(&node->bindings, network);
}
int padico_na_network_contains_host(padico_topo_network_t network, padico_na_host_t host)
{
  return na_bindings_vect_contains_network(&host->bindings, network);
}

static topo_net_entity_t na_network_getentity_byaddr(padico_topo_network_t network, const void*_addr)
{
  /*
  struct topo_binding_s key;
  void*addr = topo_binding_getaddr(&key);
  key.network = network;
  memcpy(addr, _addr, network->addr_size);
  struct topo_binding_s*binding = puk_hashtable_lookup(network->bindings, addr);
  return binding?binding->entity:(topo_net_entity_t){ .as_ptr = NULL};
  */
  /* TODO- linear search for now, since hashing variable-length addresses is broken */
#warning TODO
  topo_binding_vect_itor_t b;
  puk_vect_foreach(b, topo_binding, &network->bindings)
    {
      const struct topo_address_s*addr = topo_binding_getaddr(*b);
      if(addr && (memcmp(addr->_addr_storage, _addr, network->addr_size) == 0))
	{
	  return (*b)->entity;
	}
    }
  topo_net_entity_t e;
  e.as_ptr = NULL;
  return e;
}

void padico_na_network_gethosts(padico_topo_network_t n, padico_na_host_vect_t hosts)
{
  if(n->scope == PADICO_NA_SCOPE_HOST)
    {
      topo_binding_vect_itor_t b;
      puk_vect_foreach(b, topo_binding, &n->bindings)
	{
	  const topo_net_entity_t e = (*b)->entity;
	  if(e.kind == TOPO_ENTITY_HOST)
	    {
	      padico_na_host_vect_push_back(hosts, e.host);
	    }
	}
    }
}

void padico_na_network_get_common_host(padico_topo_network_t n1, padico_topo_network_t n2, padico_na_host_vect_t hosts)
{
  if(n1->scope == PADICO_NA_SCOPE_HOST && n2->scope == PADICO_NA_SCOPE_HOST)
    {
      topo_binding_vect_itor_t b1;
      puk_vect_foreach(b1, topo_binding, &n1->bindings)
	{
	  const topo_net_entity_t e1 = (*b1)->entity;
	  if(e1.kind == TOPO_ENTITY_HOST)
	    {
	      topo_binding_vect_itor_t b2;
	      puk_vect_foreach(b2, topo_binding, &n2->bindings)
		{
		  const topo_net_entity_t e2 = (*b2)->entity;
		  if(e2.kind == TOPO_ENTITY_HOST && e1.host == e2.host)
		    {
		      padico_na_host_vect_push_back(hosts, e1.host);
		    }
		}
	    }
	}
    }
}

void padico_na_network_set_perf(padico_topo_network_t n, int latency, int bandwidth)
{
  assert(n != NULL);
  if(latency > 0)
    n->perf.lat = latency;  
  if(bandwidth > 0)
    n->perf.bw = bandwidth;
}

/* *** lookup */

/* network lookup */

padico_topo_network_t padico_na_network_getbyid(const char*network_id)
{
  int i;
  for(i = 0; i < padico_na_network_vect_size(na_topology.networks); i++)
    {
      padico_topo_network_t network = padico_na_network_vect_at(na_topology.networks, i);
      if(strcmp(network->network_id, network_id) == 0)
	{
	  return network;
	}
    }
  return NULL;
}

padico_na_network_vect_t padico_topo_networks_getbyfamily(int family)
{
  padico_na_network_vect_t networks = padico_na_network_vect_new();
  int i;
  for(i = 0; i < padico_na_network_vect_size(na_topology.networks); i++)
    {
      padico_topo_network_t n = padico_na_network_vect_at(na_topology.networks, i);
      if(n->family == family)
	{
	  padico_na_network_vect_push_back(networks, n);
	}
    }
  return networks;
}

/* node lookup */

padico_na_node_t padico_na_network_getnodebyaddr(padico_topo_network_t network, const void*_addr)
{
  assert(network->scope == PADICO_NA_SCOPE_NODE);
  const topo_net_entity_t entity = na_network_getentity_byaddr(network, _addr);
  return entity.node;
}

padico_na_node_t padico_na_getnodebyname(const char*nodename)
{
  padico_out(50, "looking up node with name: %s\n", nodename);
  padico_na_node_t node = puk_hashtable_lookup(na_topology.nodesbyname, nodename);
  if(node)
    {
      padico_out(50, "node found: %s\n", nodename);
      return node;
    }
  padico_out(50, "no node found. Trying hosts.\n");
  padico_na_host_t host = topo_host_lookup(nodename);
  if(host)
    {
      return padico_na_host_getnode(host);
    }
  return NULL;
}

padico_na_node_t padico_na_getnodebyuuid(padico_topo_uuid_t uuid)
{
  return puk_hashtable_lookup(na_topology.node_table, uuid);
}

/* host lookup */

padico_na_host_t padico_na_network_gethostbyaddr(padico_topo_network_t network, const void*_addr)
{
  if (network->scope == PADICO_NA_SCOPE_HOST)
    {
      const topo_net_entity_t entity = na_network_getentity_byaddr(network, _addr);
      return entity.host;
    }
  return NULL;
}

padico_na_host_t padico_na_host_getbyinaddr(const struct in_addr*addr)
{
  padico_na_network_vect_t inet_networks = padico_topo_networks_getbyfamily(AF_INET);
  int i;
  for(i = 0; i < padico_na_network_vect_size(inet_networks); i++)
    {
      padico_topo_network_t net = padico_na_network_vect_at(inet_networks, i);
      const padico_na_host_t host = padico_na_network_gethostbyaddr(net, addr);
      if(host != NULL)
	return host;
    }
  return NULL;
}

padico_na_host_t padico_na_host_getbyin6addr(const struct in6_addr*addr)
{
  padico_na_network_vect_t inet_networks = padico_topo_networks_getbyfamily(AF_INET6);
  int i;
  for(i = 0; i < padico_na_network_vect_size(inet_networks); i++)
    {
      padico_topo_network_t net = padico_na_network_vect_at(inet_networks, i);
      const padico_na_host_t host = padico_na_network_gethostbyaddr(net, addr);
      if(host != NULL)
	return host;
    }
  return NULL;
}

padico_na_host_t padico_na_host_resolve_byinaddr(const struct in_addr*addr)
{
  padico_out(50, "resolving up host with addr: %s\n", inet_ntoa(*addr));
  padico_na_host_t host = padico_na_host_getbyinaddr(addr);
  if(!host)
    {
      /* no host in database with this address- perform a full resolve */
      char s_addr[INET_ADDRSTRLEN];
      inet_ntop(AF_INET, addr, s_addr, INET_ADDRSTRLEN);
      host = padico_na_host_resolvebyname(s_addr);
    }
  return host;
}

padico_na_host_t padico_na_host_resolve_byin6addr(const struct in6_addr*addr)
{
  padico_na_host_t host = padico_na_host_getbyin6addr(addr);
  if(!host)
    {
      /* no host in database with this address- perform a full resolve */
      char s_addr[INET6_ADDRSTRLEN];
      inet_ntop(AF_INET6, addr, s_addr, INET6_ADDRSTRLEN);
      host = padico_na_host_resolvebyname(s_addr);
    }
  return host;
}

/** Create a new host from host entry */
static padico_na_host_t topo_host_from_hostent(const struct hostent*he)
{
  assert(he != NULL);
  padico_na_host_t host = topo_host_lookup(he->h_name);
  if(!host)
    {
      padico_print("Topology: creating host %s from resolvable host entry.\n", he->h_name);
      host = padico_topo_host_create(he->h_name);
      /* add aliases */
      char**alias = he->h_aliases;
      while(*alias)
	{
	  padico_topo_host_add_name(host, *alias, -1);
	  alias++;
	}
      /* add 'inet' addresses */
      he = topo_gethostbyname2(host->hostname, AF_INET);
      if(he)
	{
	  const struct in_addr**addr = (const struct in_addr**)he->h_addr_list;
	  while(*addr)
	    {
	      padico_print("Topology: adding inet address %s for host %s.\n", inet_ntoa(**addr), host->hostname);
	      padico_topo_addr_visibility_t v = padico_topo_inet_visibility((*addr)->s_addr);
	      if(v == TOPO_ADDR_VISIBILITY_PUBLIC)
		{
		  padico_na_host_bind(host, na_topology.inet, *addr);
		}
	      else if (v == TOPO_ADDR_VISIBILITY_PRIVATE)
		{
		  padico_na_host_t localhost = padico_na_getlocalhost();
		  topo_binding_vect_itor_t b;
		  puk_vect_foreach(b, topo_binding, &localhost->bindings)
		    {
		      if((*b)->network != na_topology.inet)
			{
			  int cidr;
			  struct in_addr network_addr; 
			  if(padico_na_network_inet_info((*b)->network, &network_addr, &cidr) == 1)
			    {
			      assert(cidr > 0 && cidr < 32);
			      uint32_t netmask = (0xffffffff << (32 - cidr));
			      netmask = htonl(netmask);
			      if(((*addr)->s_addr & netmask) == network_addr.s_addr)
				{
				  padico_na_host_bind(host, (*b)->network, *addr);
				  break;
				}
			    }
			}
		    }
  		}
	       addr++;
	    }
	}
      /* add 'inet6' addresses */
      he = topo_gethostbyname2(host->hostname, AF_INET6);
      if(he)
	{
	  const struct in6_addr**addr = (const struct in6_addr**)he->h_addr_list;
	  while(*addr)
	    {
	      char s_addr[INET6_ADDRSTRLEN];
	      inet_ntop(AF_INET6, *addr, s_addr, INET6_ADDRSTRLEN);
	      padico_print("Topology: adding inet6 address %s for host %s.\n", s_addr, host->hostname);
	      padico_na_host_bind(host, na_topology.inet6, *addr);
	      addr++;
	    }
	}
    }
  return host;
}

/** resolve a host from name */
padico_na_host_t padico_na_host_resolvebyname(const char*hostname)
{
  const struct hostent*he = NULL;
  /* try to lookup in topology database. */
  padico_na_host_t host = topo_host_lookup(hostname);
  if(host)
    {
      return host;
    }
  /* try to interpret as IPv4 address */
  struct sockaddr_in inaddr;
  const int isv4 = inet_pton(AF_INET, hostname, &inaddr);
  if(isv4 == 1)
    {
      host = padico_na_network_gethostbyaddr(na_topology.inet, &inaddr);
      if(!host)
	{
	  he = topo_gethostbyaddr(&inaddr, sizeof(struct in_addr), AF_INET);
	  if(he)
	    {
	      host = topo_host_from_hostent(he);
	    }
	  else
	    {
	      padico_print("Topology: creating host %s as non-resolvable IPv4 address.\n", hostname);
	      host = padico_topo_host_create(hostname);
	      padico_na_host_bind(host, na_topology.inet, &inaddr.sin_addr);
	    }
	}
      return host;
    }
  /* try to interpret as IPv6 address */
  struct sockaddr_in6 in6addr;
  const int isv6 = inet_pton(AF_INET6, hostname, &in6addr);
  if(isv6 == 1)
    {
      host = padico_na_network_gethostbyaddr(na_topology.inet6, &in6addr);
      if(!host)
	{
	  he = topo_gethostbyaddr(&in6addr, sizeof(struct in6_addr), AF_INET6);
	  if(he)
	    {
	      host = topo_host_from_hostent(he);
	    }
	  else
	    {
	      padico_print("Topology: creating host %s as non-resolvable IPv6 address.\n", hostname);
	      host = padico_topo_host_create(hostname);
	      padico_na_host_bind(host, na_topology.inet6, &in6addr.sin6_addr);
	    }
	}
      return host;
    }
  /* try a DNS request with the given name */
  he = topo_gethostbyname2(hostname, AF_INET);
  if(!he)
    {
      /* host is *not* DNS-resolvable- create address-less anyway */
      int err = h_errno;
      padico_warning("Topology: hostname %s cannot be resolved (h_errno=%d, %s). Creating anyway with no address.\n",
		     hostname, err, hstrerror(err));
      host = padico_topo_host_create(hostname);
      return host;
    }
  if(strcmp(he->h_name, hostname) != 0)
    {
      host = topo_host_lookup(he->h_name);
      if(host)
	{
	  /* host was already known with an alternate name- add the name in database */
	  padico_topo_host_add_name(host, hostname, -1);
	  return host;
	}
    }
  /* create host from hostent */
  host = topo_host_from_hostent(he);
  return host;
}

/* *** Disconnection *************************************** */

void padico_na_node_disconnect(padico_na_node_t node)
{
  padico_na_node_vect_itor_t n = padico_na_node_vect_find(&node->host->nodes, node);
  if(n == NULL)
    return;
  padico_print("Topology: node lost- [%s] %s\n", node->uuid.bytes, node->canonical_nodename);
  padico_na_node_vect_erase(&node->host->nodes, n);
  topo_binding_vect_itor_t v;
  puk_vect_foreach(v, topo_binding, &node->bindings)
    {
      /*
	padico_topo_network_t network = (*v)->network;
	const struct topo_address_s*addr = topo_binding_getaddr(*v);
	puk_hashtable_remove(network->bindings, addr); 
      */ /* TODO */
      padico_free(*v);
    }
  topo_binding_vect_resize(&node->bindings, 0);
  puk_hashtable_remove(na_topology.nodesbyname, node->canonical_nodename);
  puk_hashtable_remove(na_topology.node_table, node->uuid.bytes);
}
