/** @file
 * @brief Debug tools (traces, gdb, logs, ...)
 * @ingroup Puk
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <assert.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <ctype.h>
#include <errno.h>
#include <signal.h>

#include "Puk-internals.h"
#include "Module.h"

PADICO_MODULE_HOOK(Puk);

#define TRACE_ACTION_DEFAULT 1
#define TRACE_ACTION_HIDE    2
#define TRACE_ACTION_SHOW    3
  
PUK_LIST_TYPE(puk_trace_filter,
	      int action; /**< 2: hide, 3: show */
	      int level;  /**< -1 matches all */
	      char*mod;   /**< NULL matches all */
	      char*func;  /**< NULL matches all */
	      char*file;  /**< NULL matches all */
	      int line;   /**< -1 matches all */
	      );
/** information to show on trace messages */
struct puk_trace_info_s
{
  unsigned int level:  1;
  unsigned int file:   1;
  unsigned int line:   1;
  unsigned int mod:    1;
  unsigned int func:   1;
  unsigned int date:   1;
  unsigned int plugin: 1;
};

struct puk_trace_policy_s
{
  int defaultlevel; /**< default level for traces not filtered */
  puk_trace_filter_list_t filter_list;
  struct puk_trace_info_s show; /**< what field we should show */
};

static struct
{
  /** Time origin for traces timestamps */
  struct timeval timeorig;
  /** Current trace policy */
  struct puk_trace_policy_s*trace_policy;
  /** plug-in for traces */
  puk_trace_plugin_t trace_plugin;
  /** pid of attached gdb (-1 if none) */
  int gdb_pid;
  /** enable verbosity */
  int verbose;
} puk_debug =
  {
    .trace_policy = NULL,
    .trace_plugin = NULL,
    .gdb_pid      = -1,
    .verbose      = 0
  };

/* traces policy example:
  <tracepolicy defaultlevel="10">
    <info file line mod func />
    <show mod="VIO" level="50"/>
    <show file="vnode.c" func="vsock_madw_close"/>
    <show mod="NetAccess"/>
    <hide mod="NetAccess" file="na-tcp.c"/>
  </tracepolicy>

 */

/* ********************************************************* */

void puk_trace_breakpoint(void)
{
  /* empty */
}

/******************************************************/
/* ** display functions*/

void padico_print(const char*format, ...)
{
  if(puk_debug.verbose)
    {
      va_list ap;
#define BIG_ENOUGH 4096
      /* it IS safe to use a hard-coded value. Longer messages will be 
       * truncated by snprintf/vsnprintf. Static buffer protected by puk_spinlock. */
      char buffer[BIG_ENOUGH];
      int i;
      va_start(ap, format);
      i = snprintf(buffer, BIG_ENOUGH, "%s: ", padico_hostname());
      i+= vsnprintf(&buffer[i], BIG_ENOUGH - i, format, ap);
      va_end(ap);
      fprintf(stderr, buffer);
      fflush(stderr);
    }
}

void padico_fatal(const char*format, ...)
{
  va_list ap;
  va_start(ap, format);
  fprintf(stderr, "%s: FATAL- ", padico_hostname());
  vfprintf(stderr, format, ap);
  va_end(ap);
  fprintf(stderr, "%s: abort.\n", padico_hostname());
  abort();
}

void padico_warning(const char*format, ...)
{
  va_list ap;
  va_start(ap, format);
  fprintf(stderr, "# %s: WARNING- ", padico_hostname());
  vfprintf(stderr, format, ap);
  va_end(ap);
}

/******************************************************/


static void tracepolicy_start_handler(puk_parse_entity_t e)
{ 
  const char*defaultlevel = puk_parse_getattr(e, "defaultlevel");
  struct puk_trace_policy_s*policy = padico_malloc(sizeof(struct puk_trace_policy_s));
  
  policy->filter_list = puk_trace_filter_list_new();
  memset(&policy->show, 0, sizeof(struct puk_trace_info_s));
  if(defaultlevel)
    {
      policy->defaultlevel = atoi(defaultlevel);
    }
  puk_parse_set_content(e, policy);
}
static void tracepolicy_end_handler(puk_parse_entity_t e)
{
  struct puk_trace_policy_s*policy = puk_parse_get_content(e);
  if(puk_debug.trace_policy != NULL)
    {
      padico_free(puk_debug.trace_policy);
    }
  puk_debug.trace_policy = policy;
}

static inline int attr_isset(puk_parse_entity_t e, const char*label)
{
  const char*value = puk_parse_getattr(e, label);
  return (value && (strcmp(value, "yes") == 0));
}
static void info_start_handler(puk_parse_entity_t e)
{
  struct puk_trace_info_s*ti = padico_malloc(sizeof(struct puk_trace_info_s));
  ti->level  = attr_isset(e, "level");
  ti->file   = attr_isset(e, "file");
  ti->line   = attr_isset(e, "line");
  ti->mod    = attr_isset(e, "mod");
  ti->func   = attr_isset(e, "func");
  ti->date   = attr_isset(e, "date");
  ti->plugin = attr_isset(e, "plugin");
  puk_parse_set_content(e, ti);
}
static void info_end_handler(puk_parse_entity_t e)
{
  struct puk_trace_info_s*ti = puk_parse_get_content(e);
  struct puk_trace_policy_s*policy = puk_parse_get_content(puk_parse_parent(e));
  if(ti)
    {
      memcpy(&policy->show, ti, sizeof(struct puk_trace_info_s));
      padico_free(ti);
    }
}

static void showhide_start_handler(puk_parse_entity_t e)
{   
  puk_trace_filter_t f = puk_trace_filter_new();
  const char*line  = puk_parse_getattr(e, "line");
  const char*level = puk_parse_getattr(e, "level");
  const char*mod   = puk_parse_getattr(e, "mod");
  const char*func  = puk_parse_getattr(e, "func"); 
  const char*file  = puk_parse_getattr(e, "file");
  f->mod   = mod?padico_strdup(mod):NULL;
  f->func  = func?padico_strdup(func):NULL;
  f->file  = file?padico_strdup(file):NULL;
  f->line  = line?atoi(line):-1;
  f->level = level?atoi(level):-1;
  puk_parse_set_content(e, f);
}
static void showhide_end_handler(puk_parse_entity_t e)
{
  puk_trace_filter_t f = puk_parse_get_content(e);
  struct puk_trace_policy_s*policy = puk_parse_get_content(puk_parse_parent(e));
  if(puk_parse_is(e, "show"))
    {
      f->action = TRACE_ACTION_SHOW;
    }
  else if(puk_parse_is(e, "hide"))
    {
      f->action = TRACE_ACTION_HIDE;
    }
  else
    {
      fprintf(stderr, "ooops! tag=%s\n", e->tag);
    }
  puk_trace_filter_list_push_back(policy->filter_list, f);
}

static const struct puk_tag_action_s tracepolicy_action =
{
  .xml_tag        = "tracepolicy",
  .start_handler  = &tracepolicy_start_handler,
  .end_handler    = &tracepolicy_end_handler,
  .required_level = PUK_TRUST_CONTROL
};
static const struct puk_tag_action_s info_action =
{
  .xml_tag        = "info",
  .start_handler  = &info_start_handler,
  .end_handler    = &info_end_handler,
  .required_level = PUK_TRUST_CONTROL
};
static const struct puk_tag_action_s show_action =
{
  .xml_tag        = "show",
  .start_handler  = &showhide_start_handler,
  .end_handler    = &showhide_end_handler,
  .required_level = PUK_TRUST_CONTROL
};
static const struct puk_tag_action_s hide_action =
{
  .xml_tag        = "hide",
  .start_handler  = &showhide_start_handler,
  .end_handler    = &showhide_end_handler,
  .required_level = PUK_TRUST_CONTROL
};

/* ********************************************************* */

void padico_puk_debug_init(void)
{
  if(padico_getattr("PADICO_VERBOSE"))
    puk_debug.verbose = 1;

  puk_xml_add_action(tracepolicy_action);
  puk_xml_add_action(info_action);
  puk_xml_add_action(show_action);
  puk_xml_add_action(hide_action);

#ifdef PADICO_TRACE_ON
  const char*policy_name = padico_getattr("PADICO_TRACE")?:"traces-off";
  char*policy_file = NULL;
  if(policy_name[0] != '/')
    {
      padico_string_t s_traces = padico_string_new();
      padico_string_catf(s_traces, "%s/etc/%s.xml", padico_root(), policy_name);
      policy_file = padico_strdup(padico_string_get(s_traces));
      padico_string_delete(s_traces);
    }
  else
    {
      policy_file = padico_strdup(policy_name);
    }
  padico_print("Puk: trace system enabled -- policy: %s\n", policy_file?:"(none)");
  if(policy_file != NULL)
    {
      struct puk_parse_entity_s e = puk_xml_parse_file(policy_file, PUK_TRUST_CONTROL);
      if(padico_rc_iserror(puk_parse_get_rc(&e)))
	{
	  padico_warning("Puk: parse error while parsing traces policy file (%s).\n",
			 padico_rc_gettext(puk_parse_get_rc(&e)));
	}
      padico_free(policy_file);
    }
  gettimeofday(&puk_debug.timeorig, NULL);
#else /* PADICO_TRACE_ON */
  padico_print("Puk: trace system disabled.\n");
#endif
  if(PUK_VG_RUNNING_ON_VALGRIND)
    {
      fprintf(stderr, "# Puk: running on valgrind with valgrind integration.\n");
    }
  //  puk_log_file = fopen(puk_log_filename, "w");
}

void padico_puk_debug_finalize(void)
{
  //  fclose(puk_log_file);
  if(puk_debug.gdb_pid != -1)
    {
      kill(puk_debug.gdb_pid, SIGTERM);
    }
  if(puk_debug.trace_policy)
    {
      puk_trace_filter_list_delete(puk_debug.trace_policy->filter_list);
      padico_free(puk_debug.trace_policy);
      puk_debug.trace_policy = NULL;
    }
}

void puk_trace_setplugin(puk_trace_plugin_t plugin)
{
  puk_debug.trace_plugin = plugin;
}

/* ********************************************************* */

/** @brief apply a trace filter to a context
 * @return 0: hide, 1: show
 */
int puk_trace_filter_apply(int level, puk_mod_t mod, const char*func, const char*file, int line)
{
  puk_trace_filter_itor_t fi;
  int show = TRACE_ACTION_DEFAULT;
  if(!puk_debug.trace_policy || !puk_debug.trace_policy->filter_list)
    {
      /* no filters: show nothing */
      return 0;
    }
  for(fi  = puk_trace_filter_list_begin(puk_debug.trace_policy->filter_list);
      fi != puk_trace_filter_list_end(puk_debug.trace_policy->filter_list);
      fi  = puk_trace_filter_list_next(fi))
  {
    const int matches = (((fi->level == -1)  || (level <= fi->level)) &&
			 ((fi->func == NULL) || (strcmp(func, fi->func) == 0)) &&
			 ((fi->file == NULL) || (strstr(file, fi->file) != NULL)) &&
			 ((fi->line == -1)   || (fi->line == line)) &&
			 ((fi->mod == NULL)  || (strcmp(mod->mod_name,fi->mod) == 0)) );
    if(matches)
      {
	show = fi->action;
      }
  }
  return (show == TRACE_ACTION_SHOW) || 
    ((show == TRACE_ACTION_DEFAULT) && (level <= puk_debug.trace_policy->defaultlevel));
}

void padico_out_func(int level,
		     puk_mod_t mod,
		     const char*func,
		     const char*file,
		     int line,
		     const char*format, ...)
{
  if((mod == NULL) || (func == NULL) || (file == NULL) || (format == NULL))
    {
      /* something is wrong with the args- further interpretation would segfault. */
      return;
    }
  if(puk_trace_filter_apply(level, mod, func, file, line))
    {
      char buffer[BIG_ENOUGH];
      char*current = buffer;
      int done = 0;
      va_list ap;
      va_start(ap, format);
      done += snprintf(current, BIG_ENOUGH - done, "# %s: ", padico_hostname());
      current = buffer + strlen(buffer);
      if(puk_debug.trace_policy->show.plugin)
	{
	  if(puk_debug.trace_plugin != NULL)
	    {
	      done += snprintf(current, BIG_ENOUGH - done, "#%p ", (*puk_debug.trace_plugin)());
	      current = buffer + strlen(buffer);
	    }
	}
      if(puk_debug.trace_policy->show.date)
	{
	  struct timeval now;
	  long sec, msec, usec;
	  gettimeofday(&now, NULL);
	  sec  = (long)(now.tv_sec - puk_debug.timeorig.tv_sec);
	  msec = now.tv_usec / 1000;
	  usec = now.tv_usec % 1000;
	  done += snprintf(current, BIG_ENOUGH - done, "[%ld:%3ld:%3ld] ", sec, msec, usec);
	  current = buffer+strlen(buffer);
	}
      if(puk_debug.trace_policy->show.file)
	{
	  done += snprintf(current, BIG_ENOUGH - done, "%s:%d ", file, line);
	  current = buffer+strlen(buffer);
	}
      if(puk_debug.trace_policy->show.mod && mod!= NULL)
	{
	  done += snprintf(current, BIG_ENOUGH - done, "%s: ", mod->mod_name);
	  current = buffer+strlen(buffer);
	}
      if(puk_debug.trace_policy->show.level)
	{
	  done += snprintf(current, BIG_ENOUGH - done, "[%d] ", level);
	  current = buffer+strlen(buffer);
	}
      if(puk_debug.trace_policy->show.func)
	{
	  done += snprintf(current, BIG_ENOUGH - done, "%s() ", func);
	  current = buffer+strlen(buffer);
	}
      vsnprintf(current, BIG_ENOUGH - done, format, ap);
      va_end(ap);
      buffer[BIG_ENOUGH-1] = 0;
      fputs(buffer, stderr);
    }
}

#if 0
void padico_log_func(int level,
		     puk_mod_t mod, 
		     char*function,
		     char*file,
		     int line,
		     char*format, ...)
{
  char buffer[BIG_ENOUGH];
  char*current = buffer;
  va_list ap;
  va_start(ap, format);
  snprintf(current, BIG_ENOUGH, "%s: ", padico_hostname());
  current = buffer + strlen(buffer);
  if(puk_debug.trace_policy->show.file)
    {
      snprintf(current, BIG_ENOUGH, "%s:%d ", file, line);
      current = buffer+strlen(buffer);
    }
  if(puk_debug.trace_policy->show.mod && mod!= NULL)
    {
      snprintf(current, BIG_ENOUGH, "%s: ", mod->mod_name);
      current = buffer+strlen(buffer);
    }
  if(puk_debug.trace_policy->show.level)
    {
      snprintf(current, BIG_ENOUGH, "[%d] ", level);
      current = buffer+strlen(buffer);
    }
  if(puk_debug.trace_policy->show.func)
    {
      snprintf(current, BIG_ENOUGH, "%s() ", function);
      current = buffer+strlen(buffer);
    }
  vsnprintf(current, BIG_ENOUGH, format, ap);
  buffer[BIG_ENOUGH-1] = 0;
  assert(puk_log_file != NULL);
  fputs(buffer, puk_log_file);
  va_end(ap);
}
#endif 

/* ********************************************************* */
/* *** gdb */


static void puk_debug_sigcont(int sig)
{
  /* Do nothing, but we _need_ a signal handler to work-around
   * the sigsuspend() bug in Linux 2.6.15.
   */
}

void puk_debug_gdb_attach(void)
{
  char s_padicod_bin[512];
  char s_gdbinit[512];
  char s_name[512];
  char s_pid[10];
  snprintf(s_name,        512, "gdb:%s/%s", padico_hostname(), padico_getattr("PADICO_NODE_UUID"));
  if(padico_getattr("PADICO_LEGACY"))
    {
      snprintf(s_padicod_bin, 512, padico_getattr("PADICO_LEGACY"));
    }
  else
    {
      snprintf(s_padicod_bin, 512, "%s/libexec/padico-d.bin", padico_root());
    }
  snprintf(s_gdbinit, 512, "%s/etc/Puk.gdbinit", padico_getenv("PADICO_ROOT"));
  snprintf(s_pid,      10, "%d", getpid());
  fprintf(stderr, "%s: Puk: spawning gdb...\n", padico_hostname());

  /* mask SIGCONT */
  struct sigaction oldact, act = { .sa_handler = &puk_debug_sigcont, .sa_flags = 0 };
  sigemptyset(&act.sa_mask);
  sigaction(SIGCONT, &act, &oldact);
  sigset_t set, oldset;
  sigemptyset(&set);
  sigaddset(&set, SIGCONT);
  sigprocmask(SIG_BLOCK, &set, &oldset);
  puk_debug.gdb_pid = fork();
  if(puk_debug.gdb_pid == 0)
    {
      /* restore sigmask */
      sigprocmask(SIG_SETMASK, &oldset, NULL);
      /* restore clean env */
      putenv("LD_PRELOAD=");
      putenv("LD_LIBRARY_PATH=");

      const char*xterm = padico_getenv("XTERM");
#ifdef XTERM
      if(!xterm)
	xterm = XTERM; 
#endif
      if(!xterm)
	padico_fatal("Puk: XTERM not available; cannot spawn gdb.\n");

      char*const argv[] =
	{
	  padico_strdup(xterm),
	  "-title", s_name,
	  "-e",
	  padico_strdup(padico_getenv("GDB")?:GDB),
	  "-x", s_gdbinit,
	  s_padicod_bin,
	  s_pid,
	  NULL
	};
      int rc = execv(argv[0], argv);
      if(rc)
	{
	  sleep(10);
	  padico_fatal(strerror(errno));
	}
    }
  sigemptyset(&set);
  sigaddset(&set, SIGINT);
  sigaddset(&set, SIGCONT);
  sigaddset(&set, SIGTRAP);
  sigaddset(&set, SIGIOT);
  sigwaitinfo(&set, NULL);
  sigaction(SIGCONT, &oldact, NULL);
  sigprocmask(SIG_SETMASK, &oldset, NULL);
  if(padico_getattr("PADICO_LEGACY"))
    {
      sleep(2);
    }
}

