/** @file 
 * @brief PukABI memory management
 * @ingroup PukABI
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>

#include "Puk-ABI.h"

#undef PUK_MEM_STAT

static char*puk_mem_strdup(const char*s);

static void(*puk_mem_invalidate_hook)(void*ptr, size_t size) = NULL;

void puk_abi_mem_invalidate(void*ptr, int size)
{
  if(puk_mem_invalidate_hook)
    (*puk_mem_invalidate_hook)(ptr, size);
}

/* ********************************************************* */

void puk_abi_mem_init()
{
  puk_abi_set_virtual(strdup, &puk_mem_strdup);
}

/* ********************************************************* */

#define PUK_MEM_CACHE_SIZE 64

static struct
{
  struct puk_mem_reg_s cache[PUK_MEM_CACHE_SIZE];
  puk_mem_register_t   reg;
  puk_mem_unregister_t unreg;
#ifdef PUK_MEM_STAT
  int cache_hit, cache_miss;
#endif
} puk_mem =
  {
    .reg   = NULL,
    .unreg = NULL,
#ifdef PUK_MEM_STAT
    .cache_hit = 0,
    .cache_miss = 0,
#endif
    .cache = { { .refcount = 0, .ptr = NULL, .len = 0 } }
  };

#ifdef PUK_MEM_STAT
void puk_abi_mem_stat(void)
{
  const int total = puk_mem.cache_hit + puk_mem.cache_miss;
  const int ratio = (total > 0) ? ((100 * puk_mem.cache_hit) / total) : 0;
  if(total > 0)
    fprintf(stderr, "PukABI: mem cache hit = %d; miss = %d (%d %%)\n",
	    puk_mem.cache_hit, puk_mem.cache_miss, ratio);
}
#endif


static void puk_mem_cache_flush(struct puk_mem_reg_s*r)
{
  assert(r->refcount == 0);
  (*puk_mem.unreg)(r->context, r->ptr, r->key);
  r->context = NULL;
  r->ptr = NULL;
  r->len = 0;
  r->key = 0;
  r->refcount = 0;
}

static struct puk_mem_reg_s*puk_mem_slot_lookup(void)
{
  static unsigned int victim = 0;
  int i;
  for(i = 0; i < PUK_MEM_CACHE_SIZE; i++)
    {
      struct puk_mem_reg_s*r = &puk_mem.cache[i];
      if(r->ptr == NULL && r->len == 0)
	{
	  return r;
	}
    }
  const int victim_first = victim;
  victim = (victim + 1) % PUK_MEM_CACHE_SIZE;
  struct puk_mem_reg_s*r = &puk_mem.cache[victim];
  while(r->refcount > 0 && victim != victim_first)
    {
      victim = (victim + 1) % PUK_MEM_CACHE_SIZE;
      r = &puk_mem.cache[victim];
    }
  if(r->refcount > 0)
    {
      return NULL;
    }
  puk_mem_cache_flush(r);
  return r;
}

extern void puk_mem_dump(void)
{
  int i;
  for(i = 0; i < PUK_MEM_CACHE_SIZE; i++)
    {
      struct puk_mem_reg_s*r = &puk_mem.cache[i];
      if(r->ptr != NULL)
	{
	  puk_mem_cache_flush(r);
	}
    }
}

/** Invalidate memory registration before it is freed.
 */
extern void puk_mem_invalidate(void*ptr, size_t size)
{
  if(ptr)
    {
      int i;
      for(i = 0; i < PUK_MEM_CACHE_SIZE; i++)
	{
	  struct puk_mem_reg_s*r = &puk_mem.cache[i];
	  const void*rbegin = r->ptr;
	  const void*rend   = r->ptr + r->len;
	  const void*ibegin = ptr;
	  const void*iend   = ptr + size;
	  if( (r->ptr != NULL) &&
	      ( ((ibegin >= rbegin) && (ibegin < rend)) ||
		((iend >= rbegin) && (iend < rend)) ||
		((ibegin >= rbegin) && (iend < rend)) ||
		((rbegin >= ibegin) && (rend < iend)) ) )
	    {
	      if(r->refcount > 0)
		{
		  fprintf(stderr, "PukABI: trying to invalidate registered memory still in use.\n");
		  abort();
		}
	      puk_mem_cache_flush(r);
	    }
	}
    }
}

/* ********************************************************* */

void puk_mem_set_handlers(puk_mem_register_t reg, puk_mem_unregister_t unreg)
{
  puk_spinlock_acquire();
  puk_mem.reg = reg;
  puk_mem.unreg = unreg;
  puk_mem_invalidate_hook = &puk_mem_invalidate;
  puk_spinlock_release();
}
void puk_mem_unreg(const struct puk_mem_reg_s*_r)
{
  struct puk_mem_reg_s*r = (struct puk_mem_reg_s*)_r;
  puk_spinlock_acquire();
  r->refcount--;
  if(r->refcount < 0)
    {
      fprintf(stderr, "PukABI: unbalanced registration detected.\n");
      abort();
    }
  puk_spinlock_release();
}
const struct puk_mem_reg_s*puk_mem_reg(void*context, const void*ptr, size_t len)
{
  assert(ptr != NULL);
  puk_spinlock_acquire();
  int i;
  for(i = 0; i < PUK_MEM_CACHE_SIZE; i++)
    {
      struct puk_mem_reg_s*r = &puk_mem.cache[i];
      if(context == r->context && ptr >= r->ptr && (ptr + len <= r->ptr + r->len))
	{
	  r->refcount++;
#ifdef PUK_MEM_STAT
	  puk_mem.cache_hit++;
#endif
	  puk_spinlock_release();
	  return r;
	}
    }
  struct puk_mem_reg_s*r = puk_mem_slot_lookup();
  if(r)
    {
      r->context  = context;
      r->ptr      = ptr;
      r->len      = len;
      r->refcount = 1;
      r->key      = (*puk_mem.reg)(context, ptr, len);
    }
  else
    {
      puk_abi_fatal("PukABI: rcache buffer full.\n");
    }
#ifdef PUK_MEM_STAT
  puk_mem.cache_miss++;
#endif
  puk_spinlock_release();
  return r;
}

/* ********************************************************* */

static char*puk_mem_strdup(const char*s1)
{
  char*p = NULL;
  if(s1)
    {
      const int size = strlen(s1) + 1;
      p = __puk_abi_stub__malloc__(size);
      if(!p)
	return NULL;
      memcpy(p, s1, size);
    }
  return p;
}
